from typing import Any, Dict, List, Optional, Tuple

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.container import BarContainer


def _prepare_class_data(
    task_data: List[Dict[str, Any]],
) -> Tuple[List[str], List[int]]:
    """Extracts class names and counts from task_data."""
    classes = [x["class_name"] for x in task_data]
    counts = [x["count"] for x in task_data]
    return classes, counts


def _annotate_bars(
    ax: plt.Axes, bars: BarContainer
) -> None:  # pragma: no cover
    """Adds numeric labels above bars."""
    for bar in bars:
        height = bar.get_height()
        ax.annotate(
            f"{height}",
            xy=(bar.get_x() + bar.get_width() / 2, height),
            xytext=(0, 5),
            textcoords="offset points",
            ha="center",
            va="bottom",
            fontsize=7,
        )


def plot_class_distribution(
    ax: plt.Axes, task_type: str, task_data: List[Dict[str, Any]]
) -> None:  # pragma: no cover
    """Plots a bar chart of class distribution.

    @type ax: plt.Axes
    @param ax: The axis to plot on.
    @type task_type: str
    @param task_type: The type of task.
    @type task_data: List[Dict[str, Any]]
    @param task_data: The task data to plot.
    """
    if not task_data:
        ax.axis("off")
        ax.set_title(f"{task_type} Class Distribution (None)", fontsize=12)
        return

    classes, counts = _prepare_class_data(task_data)
    num_classes = len(classes)
    bar_width = 1 / (num_classes**0.1) if num_classes else 1
    bars = ax.bar(classes, counts, width=bar_width, color="royalblue")

    # Set plot properties
    if counts:
        ax.set_ylim(top=max(counts) * 1.15)
    ax.set_title(f"{task_type} Class Distribution", fontsize=12, pad=15)
    ax.set_ylabel("Count", fontsize=12)
    ax.tick_params(
        axis="x", rotation=90 if num_classes > 5 else 45, labelsize=8
    )
    for tick in ax.get_xticklabels():  # type: ignore
        tick.set_ha("center")
    ax.margins(x=0.01)

    _annotate_bars(ax, bars)


def _prepare_heatmap_data(
    heatmap_data: Optional[List[List[float]]],
) -> np.ndarray:
    """Converts heatmap_data to a normalized NumPy array."""
    matrix = np.array(heatmap_data, dtype=np.float32)
    max_val = matrix.max()
    return matrix / max_val if max_val > 0 else matrix


def plot_heatmap(
    ax: plt.Axes,
    fig: plt.Figure,
    task_type: str,
    heatmap_data: Optional[List[List[float]]],
) -> None:  # pragma: no cover
    """ " Plots a heatmap of heatmap_data.

    @type ax: plt.Axes
    @param ax: The axis to plot on.
    @type fig: plt.Figure
    @param fig: The figure to plot on.
    @type task_type: str
    @param task_type: The type of task.
    @type heatmap_data: Optional[List[List[float]]]
    @param heatmap_data: The heatmap data to plot.
    """
    if heatmap_data is None:
        ax.axis("off")
        ax.set_title(f"{task_type} Heatmap (None)", fontsize=12)
        return

    matrix = _prepare_heatmap_data(heatmap_data)
    im = ax.imshow(matrix, cmap="viridis", extent=[0, 1, 0, 1], vmin=0, vmax=1)
    fig.colorbar(im, ax=ax, label="Relative Annotation Density")
    ax.set_title(f"{task_type} Heatmap", fontsize=12)



import json
import tarfile
from pathlib import Path
from typing import List

import typer
from rich import print
from rich.panel import Panel
from rich.pretty import Pretty
from typing_extensions import Annotated, TypeAlias

from luxonis_ml.nn_archive import Config

app = typer.Typer()


PathArgument: TypeAlias = Annotated[
    str, typer.Argument(..., help="Path to the NN Archive.")
]


@app.command()
def inspect(
    path: PathArgument,
    inputs: Annotated[
        bool, typer.Option(..., "-i", "--inputs", help="Print inputs info.")
    ] = False,
    metadata: Annotated[
        bool, typer.Option(..., "-m", "--metadata", help="Print metadata.")
    ] = False,
    outputs: Annotated[
        bool, typer.Option(..., "-o", "--outputs", help="Print outputs info.")
    ] = False,
    heads: Annotated[
        bool, typer.Option(..., "-h", "--heads", help="Print heads info.")
    ] = False,
):
    """Prints NN Archive configuration.

    If no options are provided, all info is printed.
    """

    with tarfile.open(path) as tar:
        extracted_cfg = tar.extractfile("config.json")

        if extracted_cfg is None:
            raise RuntimeError("Config JSON not found in the archive.")

        archive_config = Config(**json.loads(extracted_cfg.read().decode()))

    if not any([inputs, metadata, outputs, heads]):
        inputs = metadata = outputs = heads = True

    if metadata:
        print(
            Panel.fit(Pretty(archive_config.model.metadata), title="Metadata")
        )
    if heads:
        print(Panel.fit(Pretty(archive_config.model.heads), title="Heads"))
    if inputs:
        print(Panel.fit(Pretty(archive_config.model.inputs), title="Inputs"))
    if outputs:
        print(Panel.fit(Pretty(archive_config.model.outputs), title="Outputs"))


@app.command()
def extract(
    path: PathArgument,
    destination: Annotated[
        str,
        typer.Option(
            "-d", "--dest", help="Path where to extract the Archive."
        ),
    ] = ".",
):
    """Extracts NN Archive.

    Extracts the NN Archive to the destination path. By default, the
    Archive is extracted to the current working directory.
    """

    extract_path = Path(destination) / (Path(path).name.split(".")[0])
    extract_path.mkdir(exist_ok=True, parents=True)

    def safe_members(tar: tarfile.TarFile) -> List[tarfile.TarInfo]:
        """Filter members to prevent path traversal attacks."""
        safe_files = []
        for member in tar.getmembers():
            # Normalize path and ensure it's within the extraction folder
            if not member.name.startswith("/") and ".." not in member.name:
                safe_files.append(member)
            else:
                typer.echo(f"Skipping unsafe file: {member.name}")
        return safe_files

    with tarfile.open(path) as tf:
        for member in safe_members(tf):
            tf.extract(member, extract_path)

    typer.echo(f"Archive extracted to: {extract_path}")



"""Find Representative Images from Embeddings.

This module offers techniques to identify representative images or embeddings within a dataset.
This aids in achieving a condensed yet expressive view of your data.

Methods
=======
    - Greedy Search: Aims to find a diverse subset of images by maximizing the minimum similarity to any image outside the set.

    - K-Medoids: An adaptation of the k-means clustering algorithm, it partitions data into k clusters, each associated with a medoid.

Main Applications
=================
    - Dataset Reduction: Helps in representing large datasets with a minimal subset while retaining the essence.

    - Validation Set Creation: Identifies diverse samples for a robust validation set.

Dependencies
============
    - numpy
    - scikit-learn
    - kmedoids
    - luxonis_ml

Example
=======

    1. Greedy Search:

        # Assuming you have 'embeddings' as a numpy array of shape (num_images, embedding_dim)
        similarity_matrix = calculate_similarity_matrix(embeddings)
        desired_size = int(len(embeddings) * 0.1)
        selected_image_indices = find_representative_greedy(1-similarity_matrix, desired_size)

    2. K-Medoids:

        # to get all embeddings from qdrant:
        ids, embeddings = get_all_embeddings(qdrant_client, collection_name="mnist")
        # Assuming you have 'embeddings' as a numpy array of shape (num_images, embedding_dim)
        similarity_matrix = calculate_similarity_matrix(embeddings)
        desired_size = int(len(embeddings) * 0.1)
        selected_image_indices = find_representative_kmedoids(similarity_matrix, desired_size)
"""

from typing import List

import numpy as np
from kmedoids import KMedoids
from sklearn.metrics.pairwise import cosine_similarity

from luxonis_ml.embeddings.utils.vectordb import VectorDBAPI


def calculate_similarity_matrix(embeddings: np.ndarray) -> np.ndarray:
    return cosine_similarity(embeddings)


def find_representative_greedy(
    distance_matrix: np.ndarray, desired_size: int = 1000, seed: int = 0
) -> List[int]:
    """Find the most representative images using a greedy algorithm.
    Gready search of maximally unique embeddings.

    @type distance_matrix: np.array
    @param distance_matrix: The distance matrix to use.
    @type desired_size: int
    @param desired_size: The desired size of the representative set.
        Default is 1000.
    @type seed: int
    @param seed: The index of the seed image. Default is 0. Must be in
        the range [0, num_images-1].
    @rtype: List[int]
    @return: The indices of the representative images.
    """
    num_images = distance_matrix.shape[0]
    selected_images = set()
    selected_images.add(
        seed
    )  # If seed==0: start with the first image as a seed.

    while len(selected_images) < desired_size:
        max_distance = -1
        best_image = None

        for i in range(num_images):
            if i not in selected_images:
                # Calculate the minimum similarity to all previously selected images
                min_distance = min(
                    [distance_matrix[i, j] for j in selected_images]
                )
                if min_distance > max_distance:
                    max_distance = min_distance
                    best_image = i

        if best_image is not None:
            selected_images.add(best_image)

    return list(selected_images)


def find_representative_greedy_vectordb(
    vectordb_api: VectorDBAPI, desired_size: int = 1000, seed: int = None
) -> List[int]:
    """Find the most representative embeddings using a greedy algorithm
    with VectorDB.

    @note: Due to many requests, this function is very slow. Use
        vectordb_api.retrieve_all_embeddings() and
        find_representative_greedy() instead.
    @type vectordb_api: VectorDBAPI
    @param vectordb_api: The Vector database client instance to use for
        searches.
    @type desired_size: int
    @param desired_size: The desired size of the representative set.
        Default is 1000.
    @type seed: int
    @param seed: The ID of the seed embedding. Default is None, which
        means a random seed is chosen.
    @rtype: List[int]
    @return: The IDs of the representative embeddings.
    """
    all_ids = vectordb_api.retrieve_all_ids()

    if seed is None:
        seed = np.random.choice(all_ids)
    elif seed > len(all_ids):
        raise ValueError(f"Seed must be in the range [0, {len(all_ids) - 1}].")

    selected_embeddings = set()
    selected_embeddings.add(all_ids[seed])

    while len(selected_embeddings) < desired_size:
        max_similarity = -1
        best_embedding = None

        for embedding_id in all_ids:
            if embedding_id not in selected_embeddings:
                # Get similarities of the current embedding with the already selected embeddings
                _, scores = vectordb_api.get_similarity_scores(
                    embedding_id, list(selected_embeddings)
                )

                # Calculate the minimum similarity to all previously selected embeddings
                min_similarity = max(scores) if scores else -1

                if 1 - min_similarity > max_similarity:
                    max_similarity = min_similarity
                    best_embedding = embedding_id

        if best_embedding is not None:
            selected_embeddings.add(best_embedding)

    return list(selected_embeddings)


def find_representative_kmedoids(
    similarity_matrix: np.ndarray,
    desired_size: int = 1000,
    max_iter: int = 100,
    seed: int = None,
) -> List[int]:
    """Find the most representative images using k-medoids. K-medoids
    clustering of embeddings.

    @type similarity_matrix: np.array
    @param similarity_matrix: The similarity matrix to use.
    @type desired_size: int
    @param desired_size: The desired size of the representative set.
        Default is 1000.
    @type max_iter: int
    @param max_iter: The maximum number of iterations to use. Default is
        100.
    @type seed: int
    @param seed: The random seed to use. Default is None.
    @rtype: list
    @return: The indices of the representative images.
    """
    num_images = similarity_matrix.shape[0]
    k = min(
        desired_size, num_images
    )  # Choose 'k' as the desired size or the number of images, whichever is smaller.

    # Use k-medoids to cluster the images based on the similarity matrix.
    kmedoids_instance = KMedoids(
        n_clusters=k,
        metric="precomputed",
        init="random",
        max_iter=max_iter,
        random_state=seed,
    )
    medoid_indices = kmedoids_instance.fit(similarity_matrix).medoid_indices_

    selected_images = set(medoid_indices)
    return list(selected_images)



from pathlib import Path
from typing import Set

import pytest

from luxonis_ml.data import LuxonisLoader, LuxonisParser
from luxonis_ml.data.utils import get_task_type
from luxonis_ml.utils import environ


@pytest.mark.parametrize(
    ("url", "expected_task_types"),
    [
        (
            "COCO_people_subset.zip",
            {
                "boundingbox",
                "keypoints",
                "segmentation",
                "classification",
                "instance_segmentation",
            },
        ),
        (
            "Thermal_Dogs_and_People.v1-resize-416x416.coco.zip",
            {"boundingbox", "classification"},
        ),
        (
            "roboflow://team-roboflow/coco-128/2/coco",
            {"boundingbox", "classification"},
        ),
        (
            "Thermal_Dogs_and_People.v1-resize-416x416.voc.zip",
            {"boundingbox", "classification"},
        ),
        (
            "Thermal_Dogs_and_People.v1-resize-416x416.darknet.zip",
            {"boundingbox", "classification"},
        ),
        (
            "Thermal_Dogs_and_People.v1-resize-416x416.yolov4pytorch.zip",
            {"boundingbox", "classification"},
        ),
        (
            "Thermal_Dogs_and_People.v1-resize-416x416.mt-yolov6.zip",
            {"boundingbox", "classification"},
        ),
        (
            "Thermal_Dogs_and_People.v1-resize-416x416.createml.zip",
            {"boundingbox", "classification"},
        ),
        (
            "Thermal_Dogs_and_People.v1-resize-416x416.tensorflow.zip",
            {"boundingbox", "classification"},
        ),
        (
            "D2_Tile.png-mask-semantic.zip",
            {"segmentation", "classification"},
        ),
        (
            "Flowers_Classification.v2-raw.folder.zip",
            {"classification"},
        ),
        (
            "D2_ParkingLot.zip",
            {"boundingbox", "segmentation", "classification", "keypoints"},
        ),
        (
            "D2_ParkingLot_Native.zip",
            {
                "boundingbox",
                "instance_segmentation",
                "classification",
                "keypoints",
                "metadata/color",
                "metadata/brand",
            },
        ),
    ],
)
def test_dir_parser(
    dataset_name: str,
    url: str,
    expected_task_types: Set[str],
    storage_url: str,
    tempdir: Path,
):
    if not url.startswith("roboflow://"):
        url = f"{storage_url}/{url}"

    elif environ.ROBOFLOW_API_KEY is None:
        pytest.skip("Roboflow API key is not set")

    parser = LuxonisParser(
        url,
        dataset_name=dataset_name,
        delete_local=True,
        save_dir=tempdir,
    )
    dataset = parser.parse()
    assert len(dataset) > 0
    loader = LuxonisLoader(dataset)
    _, ann = next(iter(loader))
    task_types = {get_task_type(task) for task in ann}
    assert task_types == expected_task_types
    dataset.delete_dataset()



import json
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
from loguru import logger

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class Format(str, Enum):
    FIFTYONE = "fiftyone"
    ROBOFLOW = "roboflow"


class COCOParser(BaseParser):
    """Parses directory with COCO annotations to LDF.

    Expected formats::

        dataset_dir/
        ├── train/
        │   ├── data/
        │   │   ├── img1.jpg
        │   │   ├── img2.jpg
        │   │   └── ...
        │   └── labels.json
        ├── validation/
        │   ├── data/
        │   └── labels.json
        └── test/
            ├── data/
            └── labels.json

        This is default format returned when using FiftyOne package.

    or::

        dataset_dir/
            ├── train/
            │   ├── img1.jpg
            │   ├── img2.jpg
            │   └── ...
            │   └── _annotations.coco.json
            ├── valid/
            └── test/

        This is one of the formats that can be generated by
        U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def _detect_dataset_dir_format(
        dataset_dir: Path,
    ) -> Tuple[Optional[Format], List[str]]:
        """Checks if dataset directory structure is in FiftyOne or
        Roboflow format."""
        fiftyone_splits = ["train", "validation", "test"]
        roboflow_splits = ["train", "valid", "test"]
        if all((dataset_dir / split).exists() for split in fiftyone_splits):
            return Format.FIFTYONE, fiftyone_splits
        if all((dataset_dir / split).exists() for split in roboflow_splits):
            return Format.ROBOFLOW, roboflow_splits
        return None, []

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        if not split_path.exists():
            return None
        json_path = next(split_path.glob("*.json"), None)
        if not json_path:
            return None
        if json_path.name == "_annotations.coco.json":
            logger.info("Identified Roboflow format")
            image_dir = split_path
            return {"image_dir": image_dir, "annotation_path": json_path}
        logger.info("Identified FiftyOne format")
        dirs = [d for d in split_path.iterdir() if d.is_dir()]
        if len(dirs) != 1:
            return None
        image_dir = dirs[0]
        return {"image_dir": image_dir, "annotation_path": json_path}

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        dir_format, splits = COCOParser._detect_dataset_dir_format(dataset_dir)
        if dir_format is None:
            return False

        for split in splits:
            split_path = dataset_dir / split
            if COCOParser.validate_split(split_path) is None:
                return False
        return True

    def from_dir(
        self,
        dataset_dir: Path,
        use_keypoint_ann: bool = False,
        keypoint_ann_paths: Optional[Dict[str, str]] = None,
        split_val_to_test: bool = True,
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        dir_format, splits = COCOParser._detect_dataset_dir_format(dataset_dir)
        if dir_format is None:
            raise ValueError("Dataset is not in any expected format.")

        if dir_format is Format.ROBOFLOW:
            logger.warning(
                "Roboflow dataset format detected, following arguments won't be taken "
                "into account: ['use_keypoint_ann', 'keypoint_ann_paths', 'split_val_to_test']."
            )
        elif (
            dir_format is Format.FIFTYONE
            and use_keypoint_ann
            and not keypoint_ann_paths
        ):
            keypoint_ann_paths = {
                "train": "raw/person_keypoints_train2017.json",
                "val": "raw/person_keypoints_val2017.json",
                # NOTE: this file is not present by default
                "test": "raw/person_keypoints_test2017.json",
            }

        train_paths = COCOParser.validate_split(dataset_dir / splits[0])
        if train_paths is None:
            raise ValueError("Train split not in expected format")

        train_ann_path = (
            dataset_dir / keypoint_ann_paths["train"]
            if keypoint_ann_paths
            and use_keypoint_ann
            and dir_format is Format.FIFTYONE
            else train_paths["annotation_path"]
        )
        cleaned_annotation_path = clean_annotations(train_ann_path)
        added_train_imgs = self._parse_split(
            image_dir=train_paths["image_dir"],
            annotation_path=cleaned_annotation_path,
        )

        val_paths = COCOParser.validate_split(dataset_dir / splits[1])
        if val_paths is None:
            raise ValueError("Val split not in expected format")

        val_ann_path = (
            dataset_dir / keypoint_ann_paths["val"]
            if keypoint_ann_paths
            and use_keypoint_ann
            and dir_format is Format.FIFTYONE
            else val_paths["annotation_path"]
        )
        _added_val_imgs = self._parse_split(
            image_dir=val_paths["image_dir"], annotation_path=val_ann_path
        )

        if split_val_to_test and dir_format == Format.FIFTYONE:
            split_point = round(len(_added_val_imgs) * 0.5)
            added_val_imgs = _added_val_imgs[:split_point]
            added_test_imgs = _added_val_imgs[split_point:]
        else:
            added_val_imgs = _added_val_imgs
            added_test_imgs = []
            # NOTE: test split annotations are not included by default for FiftyOne format
            test_paths = COCOParser.validate_split(dataset_dir / splits[2])
            if test_paths is None:
                raise ValueError("Test split not in expected format")

            test_ann_path = (
                dataset_dir / keypoint_ann_paths["test"]
                if keypoint_ann_paths
                and use_keypoint_ann
                and dir_format == Format.FIFTYONE
                else test_paths["annotation_path"]
            )

            added_test_imgs = self._parse_split(
                image_dir=test_paths["image_dir"],
                annotation_path=test_ann_path,
            )

        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(
        self, image_dir: Path, annotation_path: Path
    ) -> ParserOutput:
        """Parses annotations from COCO format to LDF. Annotations
        include classification, segmentation, object detection and
        keypoints if present.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type annotation_path: Path
        @param annotation_path: Path to annotation json file
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """

        with open(annotation_path) as f:
            annotation_data = json.load(f)

        coco_images = annotation_data["images"]
        coco_annotations = annotation_data["annotations"]
        coco_categories = annotation_data["categories"]
        categories = {cat["id"]: cat["name"] for cat in coco_categories}

        skeletons = {}
        for cat in coco_categories:
            if "keypoints" in cat and "skeleton" in cat:
                skeletons[categories[cat["id"]]] = {
                    "labels": cat["keypoints"],
                    "edges": list(
                        map(tuple, (np.array(cat["skeleton"]) - 1).tolist())
                    ),
                }

        def generator() -> DatasetIterator:
            img_dict = {img["id"]: img for img in coco_images}
            ann_dict = {}
            for ann in coco_annotations:
                img_id = ann["image_id"]
                if img_id not in ann_dict:
                    ann_dict[img_id] = []
                ann_dict[img_id].append(ann)

            for img_id, img in img_dict.items():
                file: Path = image_dir.absolute().resolve() / img["file_name"]
                if not file.exists():
                    continue

                img_anns = ann_dict.get(img_id, [])
                img_h = img["height"]
                img_w = img["width"]

                for i, ann in enumerate(img_anns):
                    if ann.get("iscrowd"):
                        continue
                    class_name = categories[ann["category_id"]]

                    segmentation = None

                    coco_seg = ann.get("segmentation", [])

                    if isinstance(coco_seg, list) and coco_seg:
                        poly = []
                        for s in coco_seg:
                            poly_arr = np.array(s).reshape(-1, 2)
                            poly += [
                                (
                                    poly_arr[i, 0] / img_w,
                                    poly_arr[i, 1] / img_h,
                                )
                                for i in range(len(poly_arr))
                            ]
                        segmentation = {
                            "height": img_h,
                            "width": img_w,
                            "points": poly,
                        }
                    elif isinstance(coco_seg, dict):
                        segmentation = {
                            "height": coco_seg["size"][0],
                            "width": coco_seg["size"][1],
                            "counts": coco_seg["counts"],
                        }

                    x, y, w, h = ann["bbox"]

                    record = {
                        "file": file,
                        "annotation": {
                            "class": class_name,
                            "instance_id": i,
                            "boundingbox": {
                                "x": x / img_w,
                                "y": y / img_h,
                                "w": w / img_w,
                                "h": h / img_h,
                            },
                        },
                    }

                    if segmentation is not None:
                        record["annotation"]["segmentation"] = segmentation
                        record["annotation"]["instance_segmentation"] = (
                            segmentation
                        )

                    if "keypoints" in ann:
                        kpts = np.array(ann["keypoints"]).reshape(-1, 3)

                        np.clip(kpts[:, 0], 0, img_w, out=kpts[:, 0])
                        np.clip(kpts[:, 1], 0, img_h, out=kpts[:, 1])

                        keypoints = []
                        for kp in kpts:
                            keypoints.append(
                                (kp[0] / img_w, kp[1] / img_h, int(kp[2]))
                            )

                        record["annotation"]["keypoints"] = {
                            "keypoints": keypoints
                        }

                    yield record

        added_images = self._get_added_images(generator())

        return generator(), skeletons, added_images


def clean_annotations(annotation_path: Path) -> Path:
    """Cleans annotations by removing images that are known to cause
    issues.

    @type annotation_path: Path
    @param annotation_path: Path to the annotation JSON file.
    @rtype: Path
    @return: Path to the cleaned annotation JSON file
        ("labels_fixed.json").
    """

    files_to_avoid = [
        "000000341448.jpg",
        "000000279522.jpg",
        "000000090169.jpg",
        "000000321238.jpg",
        "000000242807.jpg",
        "000000297126.jpg",
        "000000411274.jpg",
        "000000407259.jpg",
        "000000446141.jpg",
        "000000373199.jpg",
        "000000410810.jpg",
        "000000397819.jpg",
        "000000578492.jpg",
        "000000531721.jpg",
    ]
    with open(annotation_path) as f:
        annotation_data = json.load(f)

    filtered_images = [
        img
        for img in annotation_data["images"]
        if img["file_name"] not in files_to_avoid
    ]

    if len(filtered_images) == len(annotation_data["images"]):
        return annotation_path

    filtered_image_ids = {img["id"] for img in filtered_images}
    filtered_annotations = [
        ann
        for ann in annotation_data["annotations"]
        if ann["image_id"] in filtered_image_ids
    ]

    annotation_data["images"] = filtered_images
    annotation_data["annotations"] = filtered_annotations

    cleaned_annotation_path = annotation_path.with_name("labels_fixed.json")
    with open(cleaned_annotation_path, "w") as f:
        json.dump(annotation_data, f)

    return cleaned_annotation_path



from abc import ABC, abstractmethod
from typing import Iterable, List, Mapping, Type

from luxonis_ml.typing import LoaderOutput, Params
from luxonis_ml.utils import AutoRegisterMeta, Registry

AUGMENTATION_ENGINES: Registry[Type["AugmentationEngine"]] = Registry(
    name="augmentation_engines"
)


# TODO: The engine should probably also handle normalization
# so it doesn't have to be done by injecting a normalization
# transformation to the config in LuxonisTrain.
class AugmentationEngine(
    ABC,
    metaclass=AutoRegisterMeta,
    registry=AUGMENTATION_ENGINES,
    register=False,
):
    @abstractmethod
    def __init__(
        self,
        height: int,
        width: int,
        targets: Mapping[str, str],
        n_classes: Mapping[str, int],
        config: Iterable[Params],
        keep_aspect_ratio: bool,
        is_validation_pipeline: bool,
        min_bbox_visibility: float = 0.0,
    ):
        """Initialize augmentation pipeline from configuration.

        @type height: int
        @param height: Target image height
        @type width: int
        @param width: Target image width

        @type targets: Dict[str, str]
        @param targets: Dictionary mapping task names to task types.
            Example::
                {
                    "detection/boundingbox": "bbox",
                    "detection/segmentation": "mask",
                }

        @type n_classes: Dict[str, int]
        @param n_classes: Dictionary mapping task names to the number
            of associated classes. Example::
                {
                    "cars/boundingbox": 2,
                    "cars/segmentation": 2,
                    "motorbikes/boundingbox": 1,
                }

        @type config: List[Params]
        @param config: List of dictionaries with configuration for each
            augmentation. It is up to the augmentation engine to parse
            and interpret this configuration.

        @type keep_aspect_ratio: bool
        @param keep_aspect_ratio: Whether to keep aspect ratio
        @type is_validation_pipeline: bool
        @param is_validation_pipeline: Whether this is a validation
            pipeline (in which case some augmentations are skipped)
        @type min_bbox_visibility: float
        @param min_bbox_visibility: Minimum fraction of the original bounding box that must remain visible after augmentation.
        """
        ...

    @abstractmethod
    def apply(self, data: List[LoaderOutput]) -> LoaderOutput:
        """Apply the augmentation pipeline to the data.

        @type data: List[LuxonisLoaderOutput]
        @param data: List of data to augment. The length of the list
            must be equal to the batch size.
        @rtype: LuxonisLoaderOutput
        @return: Augmented data
        """
        ...

    @property
    @abstractmethod
    def batch_size(self) -> int:
        """Getter for the batch size.

        The batch size is the number of images necessary for the
        augmentation pipeline to work in case of batch-based
        augmentations.

        For example, if the augmentation pipeline contains the MixUp
        augmentation, the batch size should be 2.

        If the pipeline requires MixUp and also Mosaic4 augmentations,
        the batch size should be 8 (2 * 4).
        """
        ...



from pathlib import Path

import pytest

from luxonis_ml.data import LuxonisParser
from luxonis_ml.data.utils.plot_utils import (
    _prepare_class_data,
    _prepare_heatmap_data,
)


@pytest.mark.parametrize("url", ["COCO_people_subset.zip"])
def test_dataset_health(
    dataset_name: str,
    url: str,
    storage_url: str,
    tempdir: Path,
):
    url = f"{storage_url}/{url}"
    dataset = LuxonisParser(
        url,
        dataset_name=dataset_name,
        delete_local=True,
        save_dir=tempdir,
    ).parse()

    statistics = dataset.get_statistics()

    class_dists = statistics["class_distributions"][""]
    tasks = [
        "segmentation",
        "keypoints",
        "instance_segmentation",
        "boundingbox",
    ]
    for task in tasks:
        task_classes = class_dists[task]
        assert len(task_classes) == 1
        assert task_classes[0]["class_name"] == "person"
        assert task_classes[0]["count"] == 145
        _prepare_class_data(task_classes)

    heatmaps = statistics["heatmaps"][""]
    annotation_types = [
        "boundingbox",
        "keypoints",
        "segmentation",
        "instance_segmentation",
    ]
    for ann_type in annotation_types:
        grid = heatmaps[ann_type]
        _prepare_heatmap_data(grid)
        assert len(grid) == 15
        for row in grid:
            assert len(row) == 15

    assert heatmaps["segmentation"] == heatmaps["instance_segmentation"]

    assert (
        sum(sum(row) for row in statistics["heatmaps"][""]["keypoints"])
        == 1169
    )
    assert (
        sum(sum(row) for row in statistics["heatmaps"][""]["boundingbox"])
        == 145
    )
    assert (
        abs(
            sum(sum(row) for row in statistics["heatmaps"][""]["segmentation"])
            - 66953
        )
        <= 2
    )
    assert (
        abs(
            sum(
                sum(row)
                for row in statistics["heatmaps"][""]["instance_segmentation"]
            )
            - 66953
        )
        <= 2
    )



import time
from pathlib import Path
from typing import List

import cv2
import numpy as np
import typer
from typing_extensions import Annotated

from luxonis_ml.data import BucketStorage, LuxonisDataset, LuxonisLoader
from luxonis_ml.data.datasets.base_dataset import DatasetIterator
from luxonis_ml.typing import Params

name = "benchmark"

DATA_DIR = Path("data/test_data")
DATA_DIR.mkdir(parents=True, exist_ok=True)


def generator(size: int) -> DatasetIterator:
    for i in range(size):
        img = np.random.randint(0, 255, (256, 256, 3), dtype=np.uint8)
        cv2.imwrite(str(DATA_DIR / f"img_{i}.jpg"), img)
        for j in range(10):
            offset = j / 50
            yield {
                "file": str(DATA_DIR / f"img_{i}.jpg"),
                "annotation": {
                    "class": "test",
                    "boundingbox": {
                        "x": 0.1,
                        "y": 0.1,
                        "w": 0.1 + offset,
                        "h": 0.1 + offset,
                    },
                    "segmentation": {
                        "points": [
                            (0.1 + offset, 0.1),
                            (0.2 + offset, 0.2),
                            (0.3 + offset, 0.3),
                        ],
                        "width": 256,
                        "height": 256,
                    },
                    "keypoints": {
                        "keypoints": [
                            (0.1 + offset, 0.1, 1),
                            (0.2 + offset, 0.2, 2),
                            (0.3 + offset, 0.3, 1),
                        ],
                    },
                },
            }


app = typer.Typer()

normal_config: List[Params] = [
    {"name": "Defocus", "params": {"p": 1}},
    {"name": "Sharpen", "params": {"p": 1}},
    {"name": "Affine", "params": {"p": 1}},
]
batched_config: List[Params] = [
    {
        "name": "MixUp",
        "params": {"p": 1, "alpha": 0.5},
    }
]


def main(
    repeat: Annotated[int, typer.Option(..., "-r", "--repeat")] = 1,
    size: Annotated[int, typer.Option(..., "-s", "--size")] = 10_000,
    no_write: Annotated[bool, typer.Option(..., "-nw", "--no-write")] = False,
    no_read: Annotated[bool, typer.Option(..., "-nr", "--no-read")] = False,
    augment: Annotated[bool, typer.Option(..., "-a", "--augment")] = False,
    batched: Annotated[bool, typer.Option(..., "-b", "--batched")] = False,
) -> None:
    if not no_write:
        avg = 0
        for _ in range(repeat):
            dataset = LuxonisDataset(
                name, delete_local=True, bucket_storage=BucketStorage.LOCAL
            )
            t = time.time()
            dataset.add(generator(size))
            dataset.make_splits()
            avg += time.time() - t
        typer.echo(f"Time to write: {avg / repeat:.2f}s")

    if not no_read:
        avg = 0
        aug_config = None
        if augment:
            aug_config = batched_config if batched else normal_config
        for _ in range(repeat):
            dataset = LuxonisDataset(name, bucket_storage=BucketStorage.LOCAL)
            loader = LuxonisLoader(
                dataset,
                height=256 if augment else None,
                width=256 if augment else None,
                augmentation_config=aug_config,
            )
            t = time.time()
            for _, _ in loader:
                pass

            avg += time.time() - t
        typer.echo(f"Time to read: {avg / repeat:.2f}s")


if __name__ == "__main__":
    typer.run(main)



from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

import yaml

from luxonis_ml.data import DatasetIterator

from .base_parser import BaseParser, ParserOutput


class YoloV6Parser(BaseParser):
    """Parses annotations from YoloV6 annotations to LDF.

    Expected format::

        dataset_dir/
        ├── images/
        │   ├── train/
        │   │   ├── img1.jpg
        │   │   ├── img2.jpg
        │   │   └── ...
        │   ├── valid/
        │   └── test/
        ├── labels/
        │   ├── train/
        │   │   ├── img1.txt
        │   │   ├── img2.txt
        │   │   └── ...
        │   ├── valid/
        │   └── test/
        └── data.yaml


    C{data.yaml} contains names of all present classes.

    This is one of the formats that can be generated by
    U{Roboflow <https://roboflow.com/>}.
    """

    @staticmethod
    def validate_split(split_path: Path) -> Optional[Dict[str, Any]]:
        label_split = split_path.parent.parent / "labels" / split_path.name
        if not split_path.exists():
            return None
        if not label_split.exists():
            return None

        labels = label_split.glob("*.txt")
        images = BaseParser._list_images(split_path)
        if not BaseParser._compare_stem_files(images, labels):
            return None
        data_yaml = split_path.parent.parent / "data.yaml"
        if not data_yaml.exists():
            return None
        return {
            "image_dir": split_path,
            "annotation_dir": label_split,
            "classes_path": data_yaml,
        }

    @staticmethod
    def validate(dataset_dir: Path) -> bool:
        for split in ["train", "valid", "test"]:
            img_split = dataset_dir / "images" / split
            if YoloV6Parser.validate_split(img_split) is None:
                return False
        return True

    def from_dir(
        self, dataset_dir: Path
    ) -> Tuple[List[Path], List[Path], List[Path]]:
        classes_path = dataset_dir / "data.yaml"
        added_train_imgs = self._parse_split(
            image_dir=dataset_dir / "images" / "train",
            annotation_dir=dataset_dir / "labels" / "train",
            classes_path=classes_path,
        )
        added_val_imgs = self._parse_split(
            image_dir=dataset_dir / "images" / "valid",
            annotation_dir=dataset_dir / "labels" / "valid",
            classes_path=classes_path,
        )
        added_test_imgs = self._parse_split(
            image_dir=dataset_dir / "images" / "test",
            annotation_dir=dataset_dir / "labels" / "test",
            classes_path=classes_path,
        )

        return added_train_imgs, added_val_imgs, added_test_imgs

    def from_split(
        self, image_dir: Path, annotation_dir: Path, classes_path: Path
    ) -> ParserOutput:
        """Parses annotations from YoloV6 format to LDF. Annotations
        include classification and object detection.

        @type image_dir: Path
        @param image_dir: Path to directory with images
        @type annotation_dir: Path
        @param annotation_dir: Path to directory with annotations
        @type classes_path: Path
        @param classes_path: Path to yaml file with classes names
        @rtype: L{ParserOutput}
        @return: Annotation generator, list of classes names, skeleton
            dictionary for keypoints and list of added images.
        """
        with open(classes_path) as f:
            classes_data = yaml.safe_load(f)
        class_names = dict(enumerate(classes_data["names"]))

        def generator() -> DatasetIterator:
            for img_path in self._list_images(image_dir):
                ann_path = annotation_dir / img_path.with_suffix(".txt").name

                with open(ann_path) as f:
                    annotation_data = f.readlines()

                for ann_line in annotation_data:
                    class_id, x_center, y_center, width, height = list(
                        ann_line.split()
                    )
                    class_name = class_names[int(class_id)]

                    yield {
                        "file": str(img_path),
                        "annotation": {
                            "class": class_name,
                            "boundingbox": {
                                "x": float(x_center) - float(width) / 2,
                                "y": float(y_center) - float(height) / 2,
                                "w": float(width),
                                "h": float(height),
                            },
                        },
                    }

        added_images = self._get_added_images(generator())

        return generator(), {}, added_images



from collections import defaultdict
from pathlib import Path
from typing import Dict

from luxonis_ml.data import (
    BucketStorage,
    LuxonisDataset,
    LuxonisLoader,
    UpdateMode,
)
from luxonis_ml.data.datasets.base_dataset import DatasetIterator
from luxonis_ml.data.utils import get_task_name, get_task_type

from .utils import create_image

STEP = 10


def compute_histogram(dataset: LuxonisDataset) -> Dict[str, int]:
    classes = defaultdict(int)
    loader = LuxonisLoader(
        dataset, exclude_empty_annotations=True, update_mode=UpdateMode.ALL
    )
    for _, record in loader:
        for task in record:
            if get_task_type(task) != "classification":
                classes[get_task_name(task)] += 1

    return dict(classes)


def test_task_ingestion(
    bucket_storage: BucketStorage, dataset_name: str, tempdir: Path
):
    dataset = LuxonisDataset(
        dataset_name,
        bucket_storage=bucket_storage,
        delete_local=True,
        delete_remote=True,
    )

    def generator1() -> DatasetIterator:
        for i in range(STEP):
            path = create_image(i, tempdir)
            yield {
                "file": str(path),
                "task_name": "animals",
                "annotation": {
                    "class": "dog",
                    "boundingbox": {"x": 0.1, "y": 0.1, "w": 0.1, "h": 0.1},
                },
            }
            yield {
                "file": str(path),
                "task_name": "animals",
                "annotation": {
                    "class": "cat",
                    "boundingbox": {"x": 0.5, "y": 0.5, "w": 0.1, "h": 0.3},
                },
            }
            yield {
                "file": str(path),
                "task_name": "landmass",
                "annotation": {
                    "class": "water",
                    "segmentation": {
                        "points": [
                            (0.1, 0.1),
                            (0.2, 0.8),
                            (0.8, 0.3),
                            (0, 0.5),
                            (0.5, 0.5),
                        ],
                        "width": 512,
                        "height": 512,
                    },
                },
            }
            yield {
                "file": str(path),
                "task_name": "landmass",
                "annotation": {
                    "class": "grass",
                    "segmentation": {
                        "points": [(0.1, 0.5), (0.6, 0.6), (0.7, 0.7)],
                        "width": 512,
                        "height": 512,
                    },
                },
            }

    dataset.add(generator1()).make_splits((1, 0, 0))

    classes = dataset.get_classes()

    assert set(classes["landmass"]) == {"water", "grass"}
    assert set(classes["animals"]) == {"dog", "cat"}

    assert compute_histogram(dataset) == {"animals": STEP, "landmass": STEP}

    def generator2() -> DatasetIterator:
        for i in range(STEP, 2 * STEP):
            path = create_image(i, tempdir)
            yield {
                "file": str(path),
                "annotation": {
                    "class": "dog",
                    "boundingbox": {"x": 0.1, "y": 0.1, "w": 0.1, "h": 0.1},
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "class": "cat",
                    "boundingbox": {"x": 0.5, "y": 0.5, "w": 0.1, "h": 0.3},
                },
            }

    dataset.add(generator2()).make_splits((1, 0, 0))
    classes = dataset.get_classes()
    assert set(classes["landmass"]) == {"background", "water", "grass"}
    assert set(classes["animals"]) == {"dog", "cat"}

    assert compute_histogram(dataset) == {
        "animals": 2 * STEP,
        "landmass": STEP,
    }

    def generator3() -> DatasetIterator:
        for i in range(2 * STEP, 3 * STEP):
            path = create_image(i, tempdir)
            yield {
                "file": str(path),
                "task_name": "animals",
                "annotation": {
                    "class": "dog",
                    "boundingbox": {"x": 0.15, "y": 0.25, "w": 0.1, "h": 0.1},
                },
            }
            yield {
                "file": str(path),
                "annotation": {
                    "class": "water",
                    "segmentation": {
                        "points": [
                            (0.1, 0.7),
                            (0.5, 0.2),
                            (0.3, 0.3),
                            (0.12, 0.45),
                        ],
                        "width": 512,
                        "height": 512,
                    },
                },
            }

    dataset.add(generator3()).make_splits((1, 0, 0))
    classes = dataset.get_classes()
    assert set(classes["landmass"]) == {"background", "water", "grass"}
    assert set(classes["animals"]) == {"dog", "cat"}

    assert compute_histogram(dataset) == {
        "animals": 3 * STEP,
        "landmass": 2 * STEP,
    }

    def generator4() -> DatasetIterator:
        for i in range(3 * STEP, 4 * STEP):
            path = create_image(i, tempdir)
            yield {
                "file": str(path),
                "task_name": "detection",
                "annotation": {
                    "class": "bike",
                    "boundingbox": {"x": 0.9, "y": 0.8, "w": 0.1, "h": 0.4},
                },
            }
            yield {
                "file": str(path),
                "task_name": "segmentation",
                "annotation": {
                    "class": "body",
                    "segmentation": {
                        "points": [
                            (0.1, 0.1),
                            (0.7, 0.5),
                            (0.3, 0.3),
                            (0.5, 0.5),
                        ],
                        "width": 512,
                        "height": 512,
                    },
                },
            }
            yield {
                "file": str(path),
                "task_name": "landmass-2",
                "annotation": {
                    "class": "water",
                    "segmentation": {
                        "points": [
                            (0.1, 0.1),
                            (0.8, 0.2),
                            (0.8, 0.9),
                            (0.1, 0.9),
                        ],
                        "width": 512,
                        "height": 512,
                    },
                },
            }

    dataset.add(generator4()).make_splits((1, 0, 0))
    classes = dataset.get_classes()

    assert set(classes["landmass"]) == {"background", "water", "grass"}
    assert set(classes["animals"]) == {"dog", "cat"}
    assert set(classes["landmass-2"]) == {"water"}
    assert set(classes["detection"]) == {"bike"}
    assert set(classes["segmentation"]) == {"body"}

    assert compute_histogram(dataset) == {
        "animals": 3 * STEP,
        "landmass": 2 * STEP,
        "landmass-2": STEP,
        "detection": STEP,
        "segmentation": STEP,
    }



import warnings
from abc import ABCMeta
from typing import (
    Callable,
    Dict,
    Generic,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

T = TypeVar("T", bound=type)


class Registry(Generic[T]):
    def __init__(self, name: str):
        """A Registry class to store and retrieve modules.

        @type name: str
        @ivar name: Name of the registry
        """
        self._module_dict: Dict[str, T] = {}
        self._name = name

    def __str__(self) -> str:
        return f"Registry('{self.name}')"

    def __repr__(self) -> str:
        return self.__str__()

    def __len__(self) -> int:
        return len(self._module_dict)

    def __getitem__(self, key: str) -> T:
        return self.get(key)

    def __setitem__(self, key: str, value: T) -> None:
        self._register(module=value, module_name=key, force=True)

    def __contains__(self, key: str) -> bool:
        return key in self._module_dict

    @property
    def name(self) -> str:
        return self._name

    def get(self, key: str) -> T:
        """Retrieves the registry record for the key.

        @type key: str
        @param key: Name of the registered item, I{e.g.} the class name
            in string format.
        @rtype: type
        @return: Corresponding class if L{key} exists
        @raise KeyError: If L{key} is not in the registry
        """
        module_cls = self._module_dict.get(key, None)
        if module_cls is None:
            raise KeyError(f"'{key}' not in the '{self.name}' registry.")
        return module_cls

    @overload
    def register_module(
        self, name: Optional[str] = ..., module: None = ..., force: bool = ...
    ) -> Callable[[T], T]: ...

    @overload
    def register_module(
        self, name: Optional[str] = ..., module: T = ..., force: bool = ...
    ) -> T: ...

    def register_module(
        self,
        name: Optional[str] = None,
        module: Optional[T] = None,
        force: bool = False,
    ) -> Union[T, Callable[[T], T]]:
        warnings.warn(
            "`register_module` is deprecated, use `register` instead.",
            stacklevel=2,
        )

        return self.register(name=name, module=module, force=force)

    @overload
    def register(
        self,
        module: None = ...,
        *,
        name: Optional[str] = ...,
        force: bool = ...,
    ) -> Callable[[T], T]: ...

    @overload
    def register(
        self,
        module: T = ...,
        *,
        name: Optional[str] = ...,
        force: bool = ...,
    ) -> None: ...

    def register(
        self,
        module: Optional[T] = None,
        *,
        name: Optional[str] = None,
        force: bool = False,
    ) -> Optional[Callable[[T], T]]:
        """Registers a module.

        Can be used as a decorator or as a normal method:

            >>> registry = Registry(name="modules")
            >>> @registry.register()
            ... class Foo:
            ...     pass
            >>> registry.get("Foo")
            <class '__main__.Foo'>
            >>> class Bar:
            ...     pass
            >>> registry.register(module=Bar)
            >>> registry.get("Bar")
            <class '__main__.Bar'>

        @type name: Optional[str]
        @param name: Name of the module. If C{None}, then use class name.
            Defaults to None.

        @type module: Optional[type]
        @param module: Module class to be registered. Defaults to None.

        @type force: bool
        @param force: Whether to override an existing class with the same name.
            Defaults to False.

        @rtype: Union[type, Callable[[type], type]]
        @return: Module class or register function if used as a decorator

        @raise KeyError: Raised if class name already exists and C{force==False}
        """

        if module is not None:
            return self._register(module=module, module_name=name, force=force)

        def wrapper(module: T) -> T:
            self._register(module=module, module_name=name, force=force)
            return module

        return wrapper

    def _register(
        self, module: T, module_name: Optional[str] = None, force: bool = False
    ) -> None:
        if module_name is None:
            module_name = module.__name__

        if not force and module_name in self._module_dict:
            existed_module = self._module_dict[module_name]
            raise KeyError(
                f"`{module_name}` already registred in `{self.name}` "
                f"registry at `{existed_module.__module__}`."
            )

        self._module_dict[module_name] = module


class AutoRegisterMeta(ABCMeta):
    """Metaclass for automatically registering modules.

    Can be set as a metaclass for abstract base classes. Then, all subclasses will be
    automatically registered under the name of the subclass.

    Example:

        >>> REGISTRY = Registry(name="modules")
        >>> class BaseClass(metaclass=AutoRegisterMeta, registry=REGISTRY):
        ...     pass
        >>> class SubClass(BaseClass):
        ...     pass
        >>> REGISTRY.get("SubClass")
        <class '__main__.SubClass'>
        >>> BaseClass.REGISTRY.get("SubClass")
        <class '__main__.SubClass'>
    """

    REGISTRY: Registry

    def __new__(
        cls,
        name: str,
        bases: Tuple[type, ...],
        attrs: Dict[str, type],
        register: bool = True,
        register_name: Optional[str] = None,
        registry: Optional[Registry] = None,
    ):
        """Automatically register the class.

        @type name: str
        @param name: Class name

        @type bases: Tuple[type, ...]
        @param bases: Base classes

        @type attrs: Dict[str, type]
        @param attrs: Class attributes

        @type register: bool
        @param register: Weather to register the class. Defaults to True.
            Should be set to False for abstract base classes.

        @type register_name: Optional[str]
        @param register_name: Name used for registration.
            If unset, the class name is used. Defaults to None.

        @type registry: Optional[Registry]
        @param registry: Registry to use for registration.
            Defaults to None. Has to be set in the base class.
        """
        new_class = super().__new__(cls, name, bases, attrs)
        if not hasattr(new_class, "REGISTRY"):
            if registry is not None:
                new_class.REGISTRY = registry
            elif register:
                raise ValueError(
                    "Registry has to be set in the base class or passed as an argument."
                )
        if register:
            registry = registry if registry is not None else new_class.REGISTRY
            registry.register(name=register_name or name, module=new_class)
        return new_class



import inspect
import warnings
from functools import wraps
from typing import Any, Callable, Dict, Literal, Optional, Set, Type

from rich.console import Console
from rich.logging import RichHandler
from rich.theme import Theme

from luxonis_ml.typing import PathType

from .environ import environ


def setup_logging(
    *,
    level: Optional[
        Literal["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
    ] = None,
    file: Optional[PathType] = None,
    **kwargs,
) -> None:  # pragma: no cover
    """Sets up global logging using loguru and rich.

    @type level: Optional[str]
    @param level: Logging level. If not set, reads from the environment
        variable C{LOG_LEVEL}. Defaults to "INFO".
    @type file: Optional[str]
    @param file: Path to the log file. If provided, logs will be saved
        to this file.
    @type kwargs: Any
    @param kwargs: Additional keyword arguments to pass to
        C{RichHandler}.
    """
    from loguru import logger

    level = level or environ.LOG_LEVEL
    if level not in {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}:
        raise ValueError(
            f"Invalid logging level: {level}. "
            "Use one of 'DEBUG', 'INFO', 'WARNING', 'ERROR', or 'CRITICAL'."
        )

    logger.remove()

    theme = Theme(
        {
            "logging.level.debug": "magenta",
            "logging.level.info": "green",
            "logging.level.warning": "yellow",
            "logging.level.error": "bold red",
            "logging.level.critical": "bold white on red",
        },
        inherit=True,
    )
    console = Console(theme=theme)
    logger.add(
        RichHandler(
            console=console,
            rich_tracebacks=True,
            tracebacks_show_locals=False,
            show_time=False,
            **kwargs,
        ),
        level=level,
        # NOTE: Needs to be a constant function to avoid
        # duplicate logging of exceptions, see
        # https://github.com/Delgan/loguru/issues/1172
        format=lambda _: "{message}",
        backtrace=False,
    )

    if file is not None:
        logger.add(
            file,
            level=level,
            format="<green>{time:YYYY-MM-DD HH:mm:ss}</green> [<level>{level}</level>] {message}",
            rotation=None,
        )

    def _custom_warning_handler(
        message: str,
        category: Type[Warning],
        filename: str,
        lineno: int,
        _file: Optional[str] = None,
        line: Optional[str] = None,
    ) -> None:
        text = warnings.formatwarning(
            message, category, filename, lineno, line
        )
        logger.warning(text)

    warnings.showwarning = _custom_warning_handler


def deprecated(
    *args: str,
    suggest: Optional[Dict[str, str]] = None,
    additional_message: Optional[str] = None,
    altogether: bool = False,
) -> Callable[[Callable], Callable]:
    """Decorator to mark a function or its parameters as deprecated.

    Example:

        >>> @deprecated("old_arg",
        ...             "another_old_arg",
        ...             suggest={"old_arg": "new_arg"},
        ...             additional_message="Usage of 'old_arg' is discouraged.")
        ...def my_func(old_arg, another_old_arg, new_arg=None):
        ...   pass
        >>> my_func("foo")
        >>> # DeprecationWarning: Argument 'old_arg'
        ... # in function `my_func` is deprecated and
        ... # will be removed in future versions.
        ... # Use 'new_arg' instead.
        ... # Usage of 'old_arg' is discouraged.

    @type args: str
    @param args: The names of the deprecated parameters.
    @type suggest: Dict[str, str]
    @param suggest: Suggested replacement parameters.
    @type additional_message: str
    @param additional_message: Additional message to display.
        If provided, it will be appended to the warning message.
    @type altogether: bool
    @param altogether: If True, the whole function is
        marked as deprecated. Defaults to False.
    """

    def decorator(func: Callable) -> Callable:
        sig = inspect.signature(func)

        @wraps(func)
        def wrapper(*f_args, **f_kwargs) -> Any:
            fname = func.__name__
            if altogether:
                msg = f"'{fname}' is deprecated and will be removed in future versions."
                if additional_message:
                    msg += f" {additional_message}"
                warnings.warn(msg, DeprecationWarning, stacklevel=2)

            if args:
                pos_arg_names = list(sig.parameters.keys())[: len(f_args)]

                for arg_name in pos_arg_names:
                    if arg_name in args:
                        _warn_deprecated(
                            arg_name, fname, suggest, additional_message
                        )

                for arg_name in f_kwargs:
                    if arg_name in args:
                        _warn_deprecated(
                            arg_name, fname, suggest, additional_message
                        )

            return func(*f_args, **f_kwargs)

        return wrapper

    return decorator


def log_once(logger: Callable[[str], None], message: str) -> None:
    """Logs a message only once.

    @type logger: Logger
    @param logger: The logger to use.
    @type message: str
    @param message: The message to log.
    """
    _cache: Set[str]

    if not hasattr(log_once, "_cache"):
        log_once._cache = set()

    if message not in log_once._cache:
        logger(message)
        log_once._cache.add(message)


def _warn_deprecated(
    arg_name: str,
    fname: str,
    suggest: Optional[Dict[str, str]],
    additional_message: Optional[str],
) -> None:
    replacement = suggest.get(arg_name) if suggest else None
    msg = (
        f"Argument '{arg_name}' in function '{fname}' "
        "is deprecated and will be removed in future versions."
    )
    if replacement:
        msg += f" Use '{replacement}' instead."
    if additional_message:
        msg += f" {additional_message}"
    warnings.warn(msg, DeprecationWarning, stacklevel=3)



import colorsys
import hashlib
import math
from collections import defaultdict
from typing import Dict, Generator, Hashable, Iterator, Mapping, Tuple

import cv2
import matplotlib.colors
import numpy as np
from bidict import bidict

from luxonis_ml.data.utils import get_task_name, task_type_iterator
from luxonis_ml.typing import HSV, RGB, Color, Labels

FONT = cv2.FONT_HERSHEY_SIMPLEX


class ColorMap(Mapping[Hashable, RGB]):
    """A mapping that assigns distinct RGB colors to hashable labels.

    The C{ColorMap} class generates and stores distinct colors for any
    hashable labels. Colors are lazily assigned upon request using a
    L{distinct_color_generator}.
    """

    def __init__(self):
        self._generator = distinct_color_generator()
        self._color_dict: Dict[Hashable, RGB] = {}

    def __getitem__(self, label: Hashable) -> RGB:
        if label not in self._color_dict:
            self._color_dict[label] = next(self._generator)
        return self._color_dict[label]

    def __iter__(self) -> Iterator[Hashable]:
        return iter(self._color_dict)

    def __len__(self) -> int:
        return len(self._color_dict)


def distinct_color_generator(stop: int = -1) -> Generator[RGB, None, None]:
    """Generate distinct RGB colors using the golden ratio.

    This generator produces a sequence of distinct colors in RGB format.
    The colors are generated by incrementing the hue by the golden ratio
    and keeping saturation and value fixed. This ensures a wide
    distribution of visually distinct colors.

    @param stop: Optional. The maximum number of colors to generate. If
        set to -1 (default), the generator will continue indefinitely.
    @type stop: int
    @yield: A tuple representing an RGB color, where each component (R,
        G, B) is an integer in the range [0, 255].
    @rtype: Generator[tuple[int, int, int], None, None]
    """
    golden_ratio = 0.618033988749895
    hue = 0.0
    i = 0
    while True:
        if i == stop:
            break
        hue = (hue + golden_ratio) % 1
        saturation = 0.8
        value = 0.95
        r, g, b = colorsys.hsv_to_rgb(hue, saturation, value)
        yield int(r * 255), int(g * 255), int(b * 255)
        i += 1


def resolve_color(color: Color) -> RGB:
    """Resolves a color to an RGB tuple.

    @type color: Color
    @param color: The color to resolve. Can be a string, an integer or a
        tuple.
    @rtype: Tuple[int, int, int]
    @return: The RGB tuple.
    """

    def _check_range(val: int) -> None:
        if val < 0 or val > 255:
            raise ValueError(f"Color value {val} is out of range [0, 255]")

    if isinstance(color, str):
        return matplotlib.colors.to_rgb(color)  # type: ignore
    if isinstance(color, int):
        _check_range(color)
        return color, color, color
    for c in color:
        _check_range(c)
    return color


def rgb_to_hsv(color: Color) -> HSV:
    """Converts an RGB color to HSV.

    @type color: Color
    @param color: The color to convert.
    @rtype: Tuple[float, float, float]
    @return: The HSV tuple.
    """
    r, g, b = resolve_color(color)
    h, s, br = colorsys.rgb_to_hsv(r / 255, g / 255, b / 255)
    return h * 360, s, br


def hsv_to_rgb(color: HSV) -> RGB:
    """Converts an HSV color to RGB.

    @type color: Tuple[float, float, float]
    @param color: The color to convert as an HSV tuple.
    @rtype: Tuple[int, int, int]
    @return: The RGB tuple.
    """
    h, s, b = color
    r, g, b = colorsys.hsv_to_rgb(h / 360, s, b)
    return int(r * 255), int(g * 255), int(b * 255)


def get_contrast_color(color: Color) -> RGB:
    """Returns a contrasting color for the given RGB color.

    @type color: Color
    @param color: The color to contrast.
    @rtype: Tuple[int, int, int]
    @return: The contrasting color.
    """

    h, s, v = rgb_to_hsv(resolve_color(color))
    h = (h + 180) % 360
    return hsv_to_rgb((h, s, v))


def str_to_rgb(string: str) -> RGB:
    """Converts a string to its unique RGB color.

    @type string: str
    @param string: The string to convert.
    @rtype: Tuple[int, int, int]
    @return: The RGB tuple.
    """
    h = int(hashlib.md5(string.encode()).hexdigest(), 16)  # noqa: S324
    r = (h & 0xFF0000) >> 16
    g = (h & 0x00FF00) >> 8
    b = h & 0x0000FF

    return r, g, b


def draw_dashed_rectangle(
    image: np.ndarray,
    pt1: Tuple[int, int],
    pt2: Tuple[int, int],
    color: Color,
    thickness: int = 1,
    dash_length: int = 10,
) -> None:
    """Draws a dashed rectangle on the image.

    @type image: np.ndarray
    @param image: The image to draw on.
    @type pt1: Tuple[int, int]
    @param pt1: The top-left corner of the rectangle.
    @type pt2: Tuple[int, int]
    @param pt2: The bottom-right corner of the rectangle.
    @type color: Color
    @param color: The color of the rectangle.
    @type thickness: int
    @param thickness: The thickness of the rectangle. Default is 1.
    @type dash_length: int
    @param dash_length: The length of the dashes. Default is 10.
    """
    x1, y1 = pt1
    x2, y2 = pt2

    def draw_dashed_line(p1: Tuple[int, int], p2: Tuple[int, int]) -> None:
        line_length = int(np.hypot(p2[0] - p1[0], p2[1] - p1[1]))
        dashes = [
            (i, i + dash_length)
            for i in range(0, line_length, 2 * dash_length)
        ]
        for start, end in dashes:
            end = min(end, line_length)
            start_point = (
                int(p1[0] + (p2[0] - p1[0]) * start / line_length),
                int(p1[1] + (p2[1] - p1[1]) * start / line_length),
            )
            end_point = (
                int(p1[0] + (p2[0] - p1[0]) * end / line_length),
                int(p1[1] + (p2[1] - p1[1]) * end / line_length),
            )
            cv2.line(
                image, start_point, end_point, resolve_color(color), thickness
            )

    draw_dashed_line((x1, y1), (x2, y1))
    draw_dashed_line((x2, y1), (x2, y2))
    draw_dashed_line((x2, y2), (x1, y2))
    draw_dashed_line((x1, y2), (x1, y1))


def draw_cross(
    img: np.ndarray,
    center: Tuple[int, int],
    size: int = 5,
    color: Color = 0,
    thickness: int = 1,
) -> None:
    """Draws a cross on the image.

    @type img: np.ndarray
    @param img: The image to draw on.
    @type center: Tuple[int, int]
    @param center: The center of the cross.
    @type size: int
    @param size: The size of the cross. Default is 5.
    @type color: Color
    @param color: The color of the cross. Default is black.
    @type thickness: int
    @param thickness: The thickness of the cross. Default is 1.
    """
    x, y = center
    color = resolve_color(color)
    cv2.line(img, (x - size, y), (x + size, y), color, thickness)
    cv2.line(img, (x, y - size), (x, y + size), color, thickness)


def create_text_image(
    text: str,
    width: int,
    height: int,
    font_size: float = 0.7,
    bg_color: Color = 255,
    text_color: Color = 0,
) -> np.ndarray:
    """Creates an image with the given text centered in the image.

    @type text: str
    @param text: The text to display.
    @type width: int
    @param width: The width of the image.
    @type height: int
    @param height: The height of the image.
    @type font_size: float
    @param font_size: The font size of the text. Default is 0.7.
    @type bg_color: Tuple[int, int, int]
    @param bg_color: The background color of the image. Default is
        white.
    @type text_color: Tuple[int, int, int]
    @param text_color: The color of the text. Default is black.
    """
    img = np.full((height, width, 3), resolve_color(bg_color), dtype=np.uint8)

    text_size = cv2.getTextSize(text, FONT, font_size, 1)[0]

    text_x = (width - text_size[0]) // 2
    text_y = (height + text_size[1]) // 2

    cv2.putText(
        img,
        text,
        (text_x, text_y),
        fontFace=FONT,
        fontScale=font_size,
        color=resolve_color(text_color),
        thickness=1,
        lineType=cv2.LINE_AA,
    )

    return img


def concat_images(
    image_dict: Dict[str, np.ndarray],
    padding: int = 10,
    label_height: int = 30,
) -> np.ndarray:
    """Concatenates images into a single image with labels.

    It will attempt to create a square grid of images.

    @type image_dict: Dict[str, np.ndarray]
    @param image_dict: A dictionary mapping image names to images.
    @type padding: int
    @param padding: The padding between images. Default is 10.
    @type label_height: int
    @param label_height: The height of the label. Default
    @rtype: np.ndarray
    @return: The concatenated image.
    """
    n_images = len(image_dict)
    n_cols = math.ceil(math.sqrt(n_images))
    n_rows = math.ceil(n_images / n_cols)

    max_h = max(img.shape[0] for img in image_dict.values())
    max_w = max(img.shape[1] for img in image_dict.values())

    cell_height = max_h + 2 * padding + label_height
    cell_width = max_w + 2 * padding

    output = np.full(
        (cell_height * n_rows, cell_width * n_cols, 3), 255, dtype=np.uint8
    )

    for idx, (name, img) in enumerate(image_dict.items()):
        i = idx // n_cols
        j = idx % n_cols

        y_start = i * cell_height
        x_start = j * cell_width

        label = create_text_image(name, cell_width, label_height)
        output[
            y_start : y_start + label_height, x_start : x_start + cell_width
        ] = label

        h, w = img.shape[:2]
        y_img = y_start + label_height + padding
        x_img = x_start + padding
        output[y_img : y_img + h, x_img : x_img + w] = img

    return output


def draw_bbox_label(
    image: np.ndarray,
    class_name: str,
    box: np.ndarray,
    color: Tuple[int, int, int],
    font_scale: float,
) -> None:
    """Draws the classname label at the top-left corner of the bounding
    box.

    @type image: np.ndarray
    @param image: The image to draw on.
    @type class_name: str
    @param class_name: The name of the class.
    @type box: np.ndarray
    @param box: The bounding box coordinates. The format is [class_id,
        x1, y1, x2, y2], where (x1, y1) is the top-left corner and (x2,
        y2) is the bottom-right corner.
    @type color: Tuple[int, int, int]
    @param color: The color of the label.
    @type font_scale: float
    @param font_scale: The scale of the font.
    """
    text = class_name

    text_size = cv2.getTextSize(text, FONT, font_scale, 1)[0]
    text_x = box[1]
    text_y = max(box[2] - 5, text_size[1])

    cv2.rectangle(
        image,
        (text_x - 2, text_y - text_size[1] - 2),
        (text_x + text_size[0] + 2, text_y + 2),
        color,
        -1,
    )
    cv2.putText(
        image,
        text,
        (text_x, text_y),
        FONT,
        font_scale,
        (255, 255, 255),
        1,
        cv2.LINE_AA,
    )


def draw_keypoint_label(
    image: np.ndarray,
    text: str,
    point: Tuple[int, int],
    size: int,
    color: Tuple[int, int, int],
    font_scale: float,
) -> None:
    """Draws a text label next to a keypoint on the image.

    @type image: np.ndarray
    @param image: The image to draw on.
    @type text: str
    @param text: The text to draw.
    @type point: Tuple[int, int]
    @param point: The coordinates of the keypoint.
    @type size: int
    @param size: The size of the keypoint.
    @type color: Tuple[int, int, int]
    @param color: The color of the text.
    @type font_scale: float
    """
    text_size = cv2.getTextSize(text, FONT, font_scale, 1)[0]
    text_x = point[0] + size + 2
    text_y = point[1] + text_size[1] // 2

    cv2.putText(
        image,
        text,
        (text_x, text_y),
        FONT,
        font_scale,
        color,
        1,
        cv2.LINE_AA,
    )


def visualize(
    image: np.ndarray,
    labels: Labels,
    classes: Dict[str, Dict[str, int]],
    blend_all: bool = False,
) -> np.ndarray:
    """Visualizes the labels on the image.

    @type image: np.ndarray
    @param image: The image to visualize.
    @type labels: Labels
    @param labels: The labels to visualize.
    @type class_names: Dict[str, List[str]]
    @param class_names: A dictionary mapping task names to a list of
        class names.
    @type blend_all: bool
    @param blend_all: Whether to blend all labels (apart from semantic
        segmentations) into a single image. This means mixing labels
        belonging to different tasks. Default is False.
    @rtype: np.ndarray
    @return: The visualized image.
    """
    h, w, _ = image.shape
    images = {"image": image}
    mappings = {task: bidict(c) for task, c in classes.items()}

    min_dimension = min(h, w)
    font_scale = max(0.25, min(1.1, 0.4 * min_dimension / 500))
    small_kp_size = max(2, round(0.002 * min_dimension))
    large_kp_size = max(4, round(0.004 * min_dimension))

    def create_mask(
        image: np.ndarray, arr: np.ndarray, task_name: str, is_instance: bool
    ) -> np.ndarray:
        mask_viz = np.zeros((h, w, 3)).astype(np.uint8)
        for i, mask in enumerate(arr):
            mask = cv2.resize(mask, (w, h), interpolation=cv2.INTER_NEAREST)
            if is_instance:
                task_classes = mappings[task_name]
                if len(bbox_classes[task_name]) > i:
                    class_id = bbox_classes[task_name][i]
                    color = str_to_rgb(task_classes.inverse[class_id])
                else:
                    color = (255, 0, 0)
                mask_viz[mask > 0] = color
            else:
                mask_viz[mask == 1] = (
                    str_to_rgb(mappings[task_name].inverse[i])
                    if (i != 0 or len(arr) == 1)
                    else (0, 0, 0)
                )

        binary_mask = (mask_viz > 0).astype(np.uint8)

        return np.where(
            binary_mask > 0,
            cv2.addWeighted(image, 0.4, mask_viz, 0.6, 0),
            image,
        )

    bbox_classes = defaultdict(list)

    for task, arr in task_type_iterator(labels, "segmentation"):
        task_name = get_task_name(task)
        image_name = task_name if task_name and not blend_all else "labels"
        images[image_name] = create_mask(
            image, arr, task_name, is_instance=False
        )

    for task, arr in task_type_iterator(labels, "boundingbox"):
        task_name = get_task_name(task)
        image_name = task_name if task_name and not blend_all else "labels"
        curr_image = images.get(image_name, image.copy())

        draw_function = cv2.rectangle

        is_sublabel = len(task.split("/")) > 2

        if is_sublabel:
            draw_function = draw_dashed_rectangle

        arr[:, [1, 3]] *= w
        arr[:, [2, 4]] *= h
        arr[:, 3] += arr[:, 1]
        arr[:, 4] += arr[:, 2]
        arr = arr.astype(int)

        for box in arr:
            class_id = int(box[0])
            bbox_classes[task_name].append(class_id)
            class_name = mappings[task_name].inverse[class_id]
            color = str_to_rgb(class_name)
            draw_function(
                curr_image,
                (box[1], box[2]),
                (box[3], box[4]),
                color,
                thickness=2,
            )
            draw_bbox_label(curr_image, class_name, box, color, font_scale)
        images[image_name] = curr_image

    for task, arr in task_type_iterator(labels, "instance_segmentation"):
        task_name = get_task_name(task)
        image_name = task_name if task_name and not blend_all else "labels"
        curr_image = images.get(image_name, image.copy())
        images[image_name] = create_mask(
            curr_image, arr, task_name, is_instance=True
        )

    for task, arr in task_type_iterator(labels, "keypoints"):
        task_name = get_task_name(task)
        image_name = task_name if task_name and not blend_all else "labels"
        curr_image = images.get(image_name, image.copy())

        task_classes = mappings[task_name]

        for i, kp in enumerate(arr):
            kp = kp.reshape(-1, 3)
            if len(bbox_classes[task_name]) > i:
                class_id = bbox_classes[task_name][i]
                color = get_contrast_color(
                    str_to_rgb(task_classes.inverse[class_id])
                )
            else:
                color = (255, 0, 0)
            for j, k in enumerate(kp):
                visibility = int(k[-1])
                if visibility == 0:
                    continue

                if visibility == 2:
                    draw_function = cv2.circle
                    size = small_kp_size
                else:
                    draw_function = draw_cross
                    size = large_kp_size

                point = (int(k[0] * w), int(k[1] * h))
                draw_function(
                    curr_image,
                    point,
                    size,
                    color=color,
                    thickness=2,
                )

                text = str(j)
                draw_keypoint_label(
                    curr_image, text, point, size, color, font_scale
                )

        images[image_name] = curr_image

    return concat_images(images)



import random
from pathlib import Path
from typing import List

import numpy as np
from pytest_subtests.plugin import SubTests

from luxonis_ml.data import LuxonisLoader
from luxonis_ml.data.datasets.base_dataset import DatasetIterator
from luxonis_ml.data.datasets.luxonis_dataset import LuxonisDataset

from .utils import create_dataset, create_image


def build_segmentation(
    polylines: List[List[float]], img_w: int, img_h: int
) -> dict:
    points = []
    for polyline in polylines:
        poly_arr = np.array(polyline).reshape(-1, 2)
        points += [
            (poly_arr[i, 0] / img_w, poly_arr[i, 1] / img_h)
            for i in range(len(poly_arr))
        ]
    return {"height": img_h, "width": img_w, "points": points}


def test_ordering_loader_background(tempdir: Path, subtests: SubTests):
    width, height = 512, 512
    segmentation = build_segmentation(
        [[0, 0, width / 2, 0, width / 2, height / 2, 0, height / 2]],
        width,
        height,
    )

    def generator(classes_list: List[str]) -> DatasetIterator:
        for i, class_name in enumerate(classes_list):
            yield {
                "file": create_image(i, tempdir),
                "annotation": {
                    "class": class_name,
                    "segmentation": segmentation,
                },
            }

    with subtests.test("implicit"):
        classes_list = ["dog", "cat", "airplane"]
        dataset = create_dataset(
            "test_segmentation_no_background", generator(classes_list)
        )

        loader = LuxonisLoader(dataset, height=height, width=width)
        for _, labels in loader:
            classification = labels["/classification"]
            assert len(classification) == len(classes_list) + 1
            assert not np.array_equal(
                classification, [1.0] + [0.0] * len(classes_list)
            )

        assert loader.classes[""] == {
            "background": 0,
            "airplane": 1,
            "cat": 2,
            "dog": 3,
        }

    with subtests.test("explicit"):
        classes_list = ["dog", "cat", "airplane", "background"]
        dataset = create_dataset(
            "test_segmentation_with_background", generator(classes_list)
        )

        loader = LuxonisLoader(
            dataset, view="train", height=height, width=width
        )
        for _, labels in loader:
            classification = labels["/classification"]
            assert len(classification) == len(classes_list)

        assert loader.classes[""] == {
            "background": 0,
            "airplane": 1,
            "cat": 2,
            "dog": 3,
        }


def test_ordering_loader_no_backgroun(tempdir: Path):
    width, height = 512, 512
    left_seg = build_segmentation(
        [[0, 0, width / 2, 0, width / 2, height, 0, height]], width, height
    )
    right_seg = build_segmentation(
        [[width / 2, 0, width, 0, width, height, width / 2, height]],
        width,
        height,
    )

    def generator() -> DatasetIterator:
        yield {
            "file": create_image(0, tempdir),
            "annotation": {"class": "dog", "segmentation": right_seg},
        }
        yield {
            "file": create_image(0, tempdir),
            "annotation": {"class": "cat", "segmentation": left_seg},
        }

    dataset = create_dataset("test_dual_class_segmentation", generator())

    loader = LuxonisLoader(dataset, view="train", height=height, width=width)
    for _, labels in loader:
        classification = labels["/classification"]
        assert len(classification) == 2

    assert loader.classes == {"": {"cat": 0, "dog": 1}}


def test_ordering_dataset(
    dataset_name: str, tempdir: Path, subtests: SubTests
):
    class_names = list("abcdefghijKLM")

    def generator(classes: List[str]) -> DatasetIterator:
        for i, class_name in enumerate(classes):
            img = create_image(i, tempdir)
            yield {
                "file": img,
                "task_name": "ordering",
                "annotation": {
                    "class": class_name,
                },
            }

    with subtests.test("no_background"):
        random.shuffle(class_names)
        dataset = create_dataset(dataset_name, generator(class_names))
        assert dataset.get_classes()["ordering"] == {
            "a": 0,
            "b": 1,
            "c": 2,
            "d": 3,
            "e": 4,
            "f": 5,
            "g": 6,
            "h": 7,
            "i": 8,
            "j": 9,
            "K": 10,
            "L": 11,
            "M": 12,
        }

    with subtests.test("with_background"):
        class_names.append("background")
        random.shuffle(class_names)
        dataset = create_dataset(dataset_name, generator(class_names))
        assert dataset.get_classes()["ordering"] == {
            "background": 0,
            "a": 1,
            "b": 2,
            "c": 3,
            "d": 4,
            "e": 5,
            "f": 6,
            "g": 7,
            "h": 8,
            "i": 9,
            "j": 10,
            "K": 11,
            "L": 12,
            "M": 13,
        }

    with subtests.test("set_classes"):
        dataset = LuxonisDataset(dataset_name)
        dataset.set_classes(["d", "b", "a", "c"])
        assert dataset.get_classes()["ordering"] == {
            "a": 0,
            "b": 1,
            "c": 2,
            "d": 3,
        }
        dataset.set_classes(["d", "b", "a", "c", "background"])
        assert dataset.get_classes()["ordering"] == {
            "background": 0,
            "a": 1,
            "b": 2,
            "c": 3,
            "d": 4,
        }
        dataset.set_classes({"a": 2, "b": 1, "c": 0, "d": 3, "background": 4})
        assert dataset.get_classes()["ordering"] == {
            "a": 2,
            "b": 1,
            "c": 0,
            "d": 3,
            "background": 4,
        }



import warnings
from collections import defaultdict
from math import prod
from typing import Any, Callable, Dict, Iterable, List, Literal, Tuple

import albumentations as A
import numpy as np
from loguru import logger
from typing_extensions import TypeAlias, override

from luxonis_ml.data.utils.task_utils import get_task_name, task_is_metadata
from luxonis_ml.typing import ConfigItem, LoaderOutput, Params

from .base_engine import AugmentationEngine
from .batch_compose import BatchCompose
from .batch_transform import BatchTransform
from .custom import TRANSFORMATIONS, LetterboxResize
from .utils import (
    postprocess_bboxes,
    postprocess_keypoints,
    postprocess_mask,
    preprocess_bboxes,
    preprocess_keypoints,
    preprocess_mask,
)

Data: TypeAlias = Dict[str, np.ndarray]
TargetType: TypeAlias = Literal[
    "array",
    "classification",
    "mask",
    "instance_mask",
    "bboxes",
    "keypoints",
    "metadata",
]


class AlbumentationConfigItem(ConfigItem):
    use_for_resizing: bool = False


class AlbumentationsEngine(AugmentationEngine, register_name="albumentations"):
    """Augmentation engine using the Albumentations library under the
    hood.

    Configuration Format
    ====================

    The configuration is a list of dictionaries, where the dictionaries
    contain the name of the transformation and optionally its parameters.
    It can also contain a boolean flag C{use_for_resizing} that indicates
    whether the transformation should be used for resizing. If no resizing
    augmentation is provided, the engine will use either L{A.Resize} or
    L{LetterboxResize} depending on the C{keep_aspect_ratio} parameter.

    The name must be either a valid name of an Albumentations
    transformation (accessible under the C{albumentations} namespace),
    or a name of a custom transformation registered in the
    L{TRANSFORMATIONS} registry.

    Example::

        [
            {
                "name": "Affine",
                "params": {
                    "rotate": 30,
                    "scale": 0.5,
                    "p": 0.3,
                },
            },
            {
                "name": "MixUp",
                "params": {
                    "alpha": [0.3, 0.7],
                    "p": 0.5,
                },
            },
            {
                "name": "CustomResize",
                "use_for_resizing": True,
            },
        ]

    Transformation Order
    ====================

    The order of transformations provided in the configuration is not
    guaranteed to be preserved. The transformations are divided into
    the following groups and are applied in the same order:

        1. batch transformations: Subclasses of L{BatchTransform}.

        2. spatial transformations: Subclasses of `A.DualTransform`.

        3. custom transformations: Subclasses of `A.BasicTransform`,
            but not subclasses of more specific base classes above.

        4. pixel transformations: Subclasses of `A.ImageOnlyTransform`.
            These transformations act only on the image.


    Supported Augmentations
    =======================

    Official Augmentations
    ----------------------

    All augmentations provided by the Albumentations library are supported.

    Supported Batch Augmentations
    -----------------------------

    MixUp
    ~~~~~

    MixUp is a data augmentation technique that blends 2 source
    images into a single image using a weight coefficient alpha.

    Mosaic4
    ~~~~~~~

    Mosaic4 transformation combines 4 images into a single image
    by placing them in a 2x2 grid.

    Augmenting Unsupported Tasks
    ============================

    Albumentations do not natively support all the tasks supported
    by Luxonis Data Format. This sections describes how
    unsupported tasks are handled.

    Note that the following applies only to officially supported
    augmentations. Custom augmentations can be implemented to handle
    arbitrary tasks.

    Classification
    --------------

    Classification tasks can be properly augmented only for multi-label
    tasks, where each class is tied to a bounding box. In such cases,
    the classes belonging to bboxes falling outside the image are removed.
    In other cases, the classification annotation is kept as is.

    Metadata
    --------

    Metadata tasks can contain arbitrary data and their semantics are
    unknown to the augmentation engine. Therefore, the only transformation
    applied to metadata is discarding metadata associated with boxes
    falling outside the image.

    Arrays
    ------

    Arrays are dealt with in the same way as metadata.
    The only transformation applied to arrays is discarding
    arrays associated with bboxes falling outside the image.

    Oriented Bounding Boxes
    -----------------------

    (Not yet implemented)

    Oriented bounding boxes are of shape (n_boxes, 5) where
    the last dimension contains the angle of the box.
    This format is not supported by Albumentations, however,
    Albumentations support angle to be part of the keypoints.
    So, the oriented bounding boxes are split into regular
    bounding boxes and a set of keypoints that represent
    the center of the bbox and contain the angle as the third coordinate.

    Both the keypoints and the bboxes are augmented separately.
    At the end, the angle is extracted from the keypoints and added
    back to the bounding boxes. The keypoints are discarded.

    Custom Augmentations
    ====================

    Custom augmentations can be implemented by creating a subclass
    of L{A.BasicTransform} and registering it in the L{TRANSFORMATIONS}
    registry.

    Possible target types that the augmentation can receive are:
        - 'image': The image. All augmentations should usually
            support this target. For subclasses of L{A.ImageOnlyTransform}
            or L{A.DualTransform} this means overriding the C{apply} method.

        - 'bboxes': Bounding boxes. For subclasses of L{A.DualTransform},
            this means overriding the L{apply_to_bboxes} method.

        - 'keypoints': Keypoints. For subclasses of L{A.DualTransform},
            this means overriding the L{apply_to_keypoints} method.

        - 'mask': Segmentation masks. For subclasses of L{A.DualTransform},
            this means overriding the L{apply_to_mask} method.

        - 'instance_mask': Instance segmentation masks.
            For subclasses of L{BatchTransform}, this means overriding
            the C{apply_to_instance_mask} method.

            Subclasses of L{A.DualTransform} do not support this target,
            instance masks are treated as regular masks instead.

            Custom augmentations can support instance masks by
            implementing their own logic for handling them and overriding
            the C{targets} property to include the C{instance_mask} target.

        - 'array': Arbitrary arrays. Can only be supported by custom
            augmentations by implementing their own logic and adding
            the C{array} target to the C{targets} property.

        - 'metadata': Metadata labels.
            Same situation as with the 'array' type.

        - 'classification': One-hot encoded multi-task classification labels.
            Same situation as with the 'array' type.

    Example::

        class CustomArrayAugmentation(A.BasicTransform):

            @property
            @override
            def targets(self) -> Dict[str, Any]:
                return {
                    "image": self.apply,
                    "array": self.apply_to_array,
                }

            def apply(self, image: np.ndarray, **kwargs) -> np.ndarray:
                ...

            def apply_to_array(
                self, array: np.ndarray, **kwargs
            ) -> np.ndarray:
                ...
    """

    def _should_skip_augmentation(
        self, config_item: Dict[str, Any], available_target_types: set
    ) -> bool:
        # Perspective issue was fixed in 1.4.19 albumentations version.
        skip_rules = {
            "keypoints": [
                "Perspective",
                "HorizontalFlip",
                "VerticalFlip",
                "Flip",
            ],
            "bboxes": ["Perspective"],
            "instance_mask": ["Perspective"],
        }
        augmentation_name = config_item["name"]
        skipped_for = [
            target_type
            for target_type, skip_list in skip_rules.items()
            if target_type in available_target_types
            and augmentation_name in skip_list
        ]
        if skipped_for:
            logger.warning(
                f"Skipping augmentation '{augmentation_name}' due to known issues for {skipped_for} target types."
            )
            return True
        return False

    @override
    def __init__(
        self,
        height: int,
        width: int,
        targets: Dict[str, str],
        n_classes: Dict[str, int],
        config: Iterable[Params],
        keep_aspect_ratio: bool = True,
        is_validation_pipeline: bool = False,
        min_bbox_visibility: float = 0.0,
    ):
        self.targets: Dict[str, TargetType] = {}
        self.target_names_to_tasks = {}
        self.n_classes = n_classes
        self.image_size = (height, width)

        for task, task_type in targets.items():
            target_name = self.task_to_target_name(task)

            if task_type == "array":
                target_type = "array"
                logger.warning(
                    "Array task detected. The 'array' task can contain "
                    "arbitrary data so it cannot be properly augmented. "
                    "The only applied transformation is discarding arrays "
                    "associated with bboxes falling outside the image."
                )

            elif task_type == "classification":
                target_type = "classification"
                logger.warning(
                    "Classification task detected. Classification tasks "
                    "can be properly augmented only for multi-label tasks, "
                    "where each class is tied to a bounding box. "
                    "In such cases, the classes belonging to bboxes falling "
                    "outside the image are removed. In other cases, "
                    "the annotation is kept as is."
                )

            elif task_is_metadata(task):
                target_type = "metadata"
                logger.warning(
                    "Metadata labels detected. Metadata labels can contain "
                    "arbitrary data so they cannot be properly augmented. "
                    "The only applied transformation is discarding metadata "
                    "associated with bboxes falling outside the image."
                )

            elif task_type == "segmentation":
                target_type = "mask"

            elif task_type == "instance_segmentation":
                target_type = "instance_mask"

            elif task_type == "boundingbox":
                target_type = "bboxes"

            elif task_type == "keypoints":
                target_type = "keypoints"

            else:
                raise ValueError(
                    f"Unsupported task type: '{task_type}'. "
                    f"Only 'array', 'classification', 'segmentation', "
                    f"'instance_segmentation', 'boundingbox', "
                    f"'keypoints', and 'metadata' are supported."
                )

            self.targets[target_name] = target_type
            self.target_names_to_tasks[target_name] = task

        # Necessary for official Albumentations transforms.
        targets_without_instance_mask = {
            target_name: target_type
            if target_type != "instance_mask"
            else "mask"
            for target_name, target_type in self.targets.items()
        }

        pixel_transforms = []
        spatial_transforms = []
        batch_transforms = []
        custom_transforms = []
        resize_transform = None

        if is_validation_pipeline:
            config = (a for a in config if a["name"] == "Normalize")

        available_target_types = set(self.targets.values())

        for config_item in config:
            if self._should_skip_augmentation(
                config_item, available_target_types
            ):
                continue

            cfg = AlbumentationConfigItem(**config_item)  # type: ignore

            transform = self.create_transformation(cfg)

            if cfg.use_for_resizing:
                logger.info(f"Using '{cfg.name}' for resizing.")
                if resize_transform is not None:
                    raise ValueError(
                        "Only one resizing augmentation can be provided."
                    )
                resize_transform = transform

            elif isinstance(transform, A.ImageOnlyTransform):
                pixel_transforms.append(transform)
            elif isinstance(transform, BatchTransform):
                batch_transforms.append(transform)
            elif isinstance(transform, A.DualTransform):
                spatial_transforms.append(transform)
            elif isinstance(transform, A.BasicTransform):
                custom_transforms.append(transform)
            else:
                raise ValueError(
                    f"Unsupported transformation type: '{transform.__name__}'. "
                    f"Only subclasses of `A.BasicTransform` are allowed. "
                )

        if resize_transform is None:
            if keep_aspect_ratio:
                resize_transform = LetterboxResize(height=height, width=width)
            else:
                resize_transform = A.Resize(height=height, width=width)

        def get_params(is_custom: bool = False) -> Dict[str, Any]:
            return {
                "bbox_params": A.BboxParams(
                    format="albumentations", min_visibility=min_bbox_visibility
                ),
                "keypoint_params": A.KeypointParams(
                    format="xy", remove_invisible=False
                ),
                "additional_targets": self.targets
                if is_custom
                else targets_without_instance_mask,
            }

        # Warning issued when "bbox_params" or "keypoint_params"
        # are provided to a compose with transformations that
        # do not use them. We don't care about these warnings.
        with warnings.catch_warnings(record=True):
            self.batch_transform = BatchCompose(
                batch_transforms, **get_params(is_custom=True)
            )
            self.spatial_transform = wrap_transform(
                A.Compose(spatial_transforms, **get_params())
            )
            self.pixel_transform = wrap_transform(
                A.Compose(pixel_transforms), is_pixel=True
            )
            self.resize_transform = wrap_transform(
                A.Compose([resize_transform], **get_params())
            )
            self.custom_transform = wrap_transform(
                A.Compose(custom_transforms, **get_params(is_custom=True))
            )

    @property
    @override
    def batch_size(self) -> int:
        return self.batch_transform.batch_size

    @override
    def apply(self, input_batch: List[LoaderOutput]) -> LoaderOutput:
        data_batch, n_keypoints = self.preprocess_batch(input_batch)

        data = self.batch_transform(data_batch)

        for target_name in list(data.keys()):
            if data[target_name].size == 0:
                del data[target_name]

        data = self.spatial_transform(**data)
        data = self.custom_transform(**data)

        transformed_size = data["image"].shape[:2]

        if transformed_size != self.image_size:
            transformed_size = prod(transformed_size)
            target_size = prod(self.image_size)

            if transformed_size > target_size:
                data = self.resize_transform(**data)
                data = self.pixel_transform(**data)
            else:
                data = self.pixel_transform(**data)
                data = self.resize_transform(**data)
        else:
            data = self.pixel_transform(**data)

        return self.postprocess(data, n_keypoints)

    def preprocess_batch(
        self, labels_batch: List[LoaderOutput]
    ) -> Tuple[List[Data], Dict[str, int]]:
        """Preprocess a batch of labels.

        @type labels_batch: List[Data]
        @param labels_batch: List of dictionaries mapping task names to
            the annotations as C{np.ndarray}
        @rtype: Tuple[List[Data], Dict[str, int]]
        @return: Tuple containing the preprocessed data and a dictionary
            mapping task names to the number of keypoints for that task.
        """
        data_batch = []
        bbox_counters = defaultdict(int)
        n_keypoints = {}

        for image, labels in labels_batch:
            data = {"image": image}
            height, width, _ = image.shape
            for target_name, target_type in self.targets.items():
                task = self.target_names_to_tasks[target_name]

                if task not in labels:
                    if target_type == "mask":
                        data[target_name] = np.empty(
                            (
                                0,
                                0,
                                self.n_classes[
                                    self.target_names_to_tasks[target_name]
                                ],
                            )
                        )
                    elif target_type == "classification":
                        data[target_name] = np.zeros(
                            self.n_classes[
                                self.target_names_to_tasks[target_name]
                            ]
                        )
                    else:
                        data[target_name] = np.array([])
                    continue

                array = labels[task]

                if target_type in {"mask", "instance_mask"}:
                    data[target_name] = preprocess_mask(array)

                elif target_type == "bboxes":
                    data[target_name] = preprocess_bboxes(
                        array, bbox_counters[target_name]
                    )
                    bbox_counters[target_name] += data[target_name].shape[0]

                elif target_type == "keypoints":
                    n_keypoints[target_name] = array.shape[1] // 3
                    data[target_name] = preprocess_keypoints(
                        array, height, width
                    )
                else:
                    data[target_name] = array

            data_batch.append(data)

        return data_batch, n_keypoints

    def postprocess(
        self, data: Data, n_keypoints: Dict[str, int]
    ) -> LoaderOutput:
        """Postprocess the augmented data back to LDF format.

        Discards labels associated with bboxes that are outside the
        image.

        @type data: Data
        @param data: Dictionary mapping task names to the annotations as
            C{np.ndarray}
        @type n_keypoints: Dict[str, int]
        @param n_keypoints: Dictionary mapping task names to the number
            of keypoints for that task.
        @rtype: LoaderOutput
        @return: Tuple containing the augmented image and the labels.
        """
        out_labels = {}
        out_image = data.pop("image")
        image_height, image_width, _ = out_image.shape

        bboxes_indices = {}

        for target_name, target_type in self.targets.items():
            if target_name not in data:
                continue

            array = data[target_name]
            if array.size == 0:
                continue

            task = self.target_names_to_tasks[target_name]
            task_name = get_task_name(task)

            if target_type == "bboxes":
                out_labels[task], index = postprocess_bboxes(array)
                bboxes_indices[task_name] = index

        for target_name, target_type in self.targets.items():
            if target_name not in data:
                continue

            array = data[target_name]
            if array.size == 0:
                continue

            task = self.target_names_to_tasks[target_name]
            task_name = get_task_name(task)

            if task_name not in bboxes_indices:
                if "bboxes" in self.targets.values():
                    bbox_ordering = np.array([], dtype=int)
                elif target_type == "keypoints":
                    bbox_ordering = np.arange(
                        array.shape[0] // n_keypoints[target_name]
                    )
                else:
                    bbox_ordering = np.arange(array.shape[0])
            else:
                bbox_ordering = bboxes_indices[task_name]

            if target_type == "mask":
                out_labels[task] = postprocess_mask(array)

            elif target_type == "instance_mask":
                mask = postprocess_mask(array)
                out_labels[task] = mask[bbox_ordering]

            elif target_type == "keypoints":
                out_labels[task] = postprocess_keypoints(
                    array,
                    bbox_ordering,
                    image_height,
                    image_width,
                    n_keypoints[target_name],
                )
            elif target_type in {"array", "metadata"}:
                out_labels[task] = array[bbox_ordering]

            elif target_type == "classification":
                out_labels[task] = array

        return out_image, out_labels

    @staticmethod
    def create_transformation(
        config: AlbumentationConfigItem,
    ) -> A.BasicTransform:
        if hasattr(A, config.name):
            return getattr(A, config.name)(**config.params)
        return TRANSFORMATIONS.get(config.name)(**config.params)  # type: ignore

    @staticmethod
    def task_to_target_name(task: str) -> str:
        target = task.replace("/", "_").replace("-", "_")
        assert target.isidentifier()
        return target


def wrap_transform(
    transform: A.BaseCompose, is_pixel: bool = False
) -> Callable[..., Data]:
    def apply_transform(**data: np.ndarray) -> Data:
        if not transform.transforms:
            return data

        if is_pixel:
            data["image"] = transform(image=data["image"])["image"]
            return data

        return transform(**data)

    return apply_transform



import json
import os
import time
from functools import wraps
from importlib.util import find_spec
from pathlib import Path
from typing import Any, Callable, Dict, Literal, Optional, Union

import cv2
import numpy as np
from loguru import logger
from unique_names_generator import get_random_name

from luxonis_ml.typing import PathType
from luxonis_ml.utils.filesystem import LuxonisFileSystem


class LuxonisTracker:
    def __init__(
        self,
        project_name: Optional[str] = None,
        project_id: Optional[str] = None,
        run_name: Optional[str] = None,
        run_id: Optional[str] = None,
        save_directory: PathType = "output",
        is_tensorboard: bool = False,
        is_wandb: bool = False,
        is_mlflow: bool = False,
        is_sweep: bool = False,
        wandb_entity: Optional[str] = None,
        mlflow_tracking_uri: Optional[str] = None,
        rank: int = 0,
    ):
        """Implementation of PytorchLightning Logger that wraps various
        logging software. Supported loggers: TensorBoard, WandB and
        MLFlow.

        @type project_name: Optional[str]
        @param project_name: Name of the project used for WandB and MLFlow.
            Defaults to None.

        @type project_id: Optional[str]
        @param project_id: Project id used for WandB and MLFlow.
            Defaults to None.

        @type run_name: Optional[str]
        @param run_name: Name of the run, if None then auto-generate random name.
            Defaults to None.

        @type run_id: Optional[str]
        @param run_id: Run id used for continuing MLFlow run.
            Defaults to None.

        @type save_directory: str
        @param save_directory: Path to save directory.
            Defaults to "output".

        @type is_tensorboard: bool
        @param is_tensorboard: Wheter use TensorBoard logging.
            Defaults to False.

        @type is_wandb: bool
        @param is_wandb: Wheter use WandB logging.
            Defaults to False.

        @type is_mlflow: bool
        @param is_mlflow: Wheter use MLFlow logging.
            Defaults to False.

        @type is_sweep: bool
        @param is_sweep: Wheter current run is part of a sweep.
            Defaults to False.

        @type wandb_entity: Optional[str]
        @param wandb_entity: WandB entity to use.
            Defaults to None.

        @type mlflow_tracking_uri: Optional[str]
        @param mlflow_tracking_uri: MLFlow tracking uri to use.
            Defaults to None.

        @type rank: int
        @param rank: Rank of the process, used when running on multiple threads.
            Defaults to 0.
        """
        os.environ["MLFLOW_HTTP_REQUEST_MAX_RETRIES"] = "2"

        self.project_name = project_name
        self.project_id = project_id
        self.save_directory = Path(save_directory)
        self.save_directory.mkdir(parents=True, exist_ok=True)
        self.is_tensorboard = is_tensorboard
        self.is_wandb = is_wandb
        self.is_mlflow = is_mlflow
        self.is_sweep = is_sweep
        self.rank = rank
        self.local_logs = {
            "metric": [],
            "params": {},
            "images": [],
            "artifacts": [],
            "matrices": [],
            "metrics": [],
        }
        self.mlflow_initialized = False

        self.run_id = (
            run_id  # if using MLFlow then it will continue previous run
        )

        if (
            (is_wandb or is_mlflow)
            and self.project_name is None
            and self.project_id is None
        ):
            raise ValueError(
                "Either project_name or project_id must be specified!"
            )

        if self.is_wandb and wandb_entity is None:
            raise ValueError("Must specify wandb_entity when using wandb!")
        self.wandb_entity = wandb_entity
        if self.is_mlflow:
            if mlflow_tracking_uri is None:
                raise ValueError(
                    "Must specify mlflow_tracking_uri when using mlflow!"
                )
            self.mlflow_tracking_uri = mlflow_tracking_uri

        if not (self.is_tensorboard or self.is_wandb or self.is_mlflow):
            raise ValueError("At least one integration must be used!")

        self._experiment = None

        if run_name:
            self.run_name = run_name
        # create new directory if rank==0 else return newest run
        elif rank == 0:
            self.run_name = self._get_run_name()
        else:
            time.sleep(1)  # DDP hotfix
            self.run_name = self._get_latest_run_name()

        self.run_directory = self.save_directory / self.run_name
        self.run_directory.mkdir(parents=True, exist_ok=True)

    @staticmethod
    def rank_zero_only(fn: Callable) -> Callable:
        """Function wrapper that lets only processes with rank=0 execute
        it."""

        @wraps(fn)
        def wrapped_fn(
            self: "LuxonisTracker", *args: Any, **kwargs: Any
        ) -> Optional[Any]:
            if self.rank == 0:
                return fn(self, *args, **kwargs)
            return None

        return wrapped_fn

    def log_to_mlflow(self, log_fn: Callable, *args, **kwargs) -> None:
        """Attempts to log to MLflow, with retries.

        Logs locally if failures persist.
        """
        try:
            log_fn(*args, **kwargs)
            self.log_stored_logs_to_mlflow()  # Attempt to log stored logs after successful log
        except Exception as e:
            logger.warning(f"Attempt to log to MLflow failed: {e}")
        else:
            return

        self.store_log_locally(log_fn, *args, **kwargs)

    def store_log_locally(self, log_fn: Callable, *args, **kwargs) -> None:
        """Stores log data locally if logging to MLflow fails."""
        # Checking functions without triggering reconnections.
        if log_fn == self.experiment["mlflow"].log_metric:
            self.local_logs["metric"].append(
                {"name": args[0], "value": args[1], "step": args[2]}
            )
        elif log_fn == self.experiment["mlflow"].log_metrics:
            self.local_logs["metrics"].append(
                {"metrics": args[0], "step": args[1]}
            )
        elif log_fn == self.experiment["mlflow"].log_params:
            self.local_logs["params"].update(args[0])
        elif log_fn == self.experiment["mlflow"].log_image:
            self.local_logs["images"].append(
                {"image_data": args[0], "name": args[1]}
            )
        elif log_fn == self.upload_artifact_to_mlflow:
            self.local_logs["artifacts"].append(
                {"path": str(args[0]), "name": args[1]}
            )
        elif log_fn == self.experiment["mlflow"].log_dict:
            self.local_logs["matrices"].append(
                {"matrix": args[0], "name": args[1]}
            )

    def log_stored_logs_to_mlflow(self) -> None:
        """Attempts to log any data stored in local_logs to MLflow."""
        if not self.mlflow_initialized or not any(self.local_logs.values()):
            return

        try:
            if self.local_logs["params"]:
                self.experiment["mlflow"].log_params(self.local_logs["params"])
                self.local_logs["params"] = {}
            for metric in list(self.local_logs["metric"]):
                self.experiment["mlflow"].log_metric(
                    metric["name"], metric["value"], metric["step"]
                )
                self.local_logs["metric"].remove(metric)
            for metrics in list(self.local_logs["metrics"]):
                self.experiment["mlflow"].log_metrics(
                    metrics["metrics"], metrics["step"]
                )
                self.local_logs["metrics"].remove(metrics)
            for image in list(self.local_logs["images"]):
                self.experiment["mlflow"].log_image(
                    image["image_data"], image["name"]
                )
                self.local_logs["images"].remove(image)
            for matrix in list(self.local_logs["matrices"]):
                self.experiment["mlflow"].log_dict(
                    matrix["matrix"], matrix["name"]
                )
                self.local_logs["matrices"].remove(matrix)
            for artifact in list(self.local_logs["artifacts"]):
                self.upload_artifact_to_mlflow(
                    Path(artifact["path"]), artifact["name"]
                )
                self.local_logs["artifacts"].remove(artifact)

            logger.info("Successfully re-logged stored logs to MLflow.")
        except Exception as e:
            logger.warning(f"Failed to re-log stored logs to MLflow: {e}")

    def save_logs_locally(self) -> None:
        """Saves metrics, parameters, images, artifacts, and matrices
        locally."""
        image_dir = self.run_directory / "images"
        artifact_dir = self.run_directory / "artifacts"

        image_dir.mkdir(exist_ok=True)
        artifact_dir.mkdir(exist_ok=True)

        for idx, img in enumerate(self.local_logs["images"]):
            img_path = str(image_dir / f"{idx}.png")
            cv2.imwrite(
                img_path, cv2.cvtColor(img["image_data"], cv2.COLOR_RGB2BGR)
            )
            img["image_data"] = img_path  # Replace data with path

        for artifact in self.local_logs["artifacts"]:
            artifact_path = Path(artifact["path"])
            if artifact_path.exists():
                local_path = artifact_dir / artifact_path.name
                local_path.write_bytes(artifact_path.read_bytes())
                artifact["path"] = str(local_path)

        log_dir = self.run_directory / "local_logs.json"
        with open(log_dir, "w") as f:
            json.dump(
                {
                    k: self.local_logs[k]
                    for k in [
                        "metrics",
                        "metric",
                        "params",
                        "images",
                        "artifacts",
                        "matrices",
                    ]
                },
                f,
            )

        logger.info(
            f"Logs saved locally at '{log_dir}', "
            f"images in {image_dir}, artifacts in {artifact_dir}"
        )

    @property
    def name(self) -> str:
        """Returns run name.

        @type: str
        """
        return self.run_name

    @property
    def version(self) -> int:
        """Returns tracker's version.

        @type: int
        """
        return 1

    @property
    @rank_zero_only
    def experiment(
        self,
    ) -> Dict[Literal["tensorboard", "wandb", "mlflow"], Any]:
        """Creates new experiments or returns active ones if already
        created."""
        if self._experiment is None:
            self._experiment = {}

        if self.is_tensorboard and "tensorboard" not in self._experiment:
            from torch.utils.tensorboard.writer import SummaryWriter

            log_dir = self.save_directory / "tensorboard_logs" / self.run_name
            if self.is_sweep:
                trial_id = 0
                if log_dir.exists():
                    trial_id = (
                        max(
                            (
                                int(f.split("_")[-1])
                                for f in os.listdir(log_dir)
                                if f.startswith("trial_")
                            ),
                            default=0,
                        )
                        + 1
                    )
                log_dir = log_dir / f"trial_{trial_id}"

            self._experiment["tensorboard"] = SummaryWriter(log_dir=log_dir)

        if self.is_wandb and "wandb" not in self._experiment:
            import wandb

            self._experiment["wandb"] = wandb

            log_dir = self.save_directory / "wandb_logs"
            log_dir.mkdir(parents=True, exist_ok=True)

            self._experiment["wandb"].init(
                project=self.project_name
                if self.project_name is not None
                else self.project_id,
                entity=self.wandb_entity,
                dir=log_dir,
                name=self.run_name,
            )

        if self.is_mlflow and self.mlflow_initialized is False:
            try:
                import mlflow

                if find_spec("psutil") is not None:
                    mlflow.enable_system_metrics_logging()
                    if find_spec("pynvml") is None:
                        logger.warning(
                            "pynvml not found, GPU stats will not be monitored. "
                            "To enable GPU monitoring, install it using 'pip install pynvml'"
                        )
                else:
                    logger.warning(
                        "`psutil` not found. To enable system metric logging, "
                        "install it using 'pip install psutil'"
                    )

                self._experiment["mlflow"] = mlflow

                self.artifacts_dir = self.run_directory / "artifacts"
                self.artifacts_dir.mkdir(parents=True, exist_ok=True)

                self._experiment["mlflow"].set_tracking_uri(
                    self.mlflow_tracking_uri
                )

                if self.project_id is not None:
                    self.project_name = None

                experiment = self._experiment["mlflow"].set_experiment(
                    experiment_name=self.project_name,
                    experiment_id=self.project_id,
                )
                self.project_id = experiment.experiment_id

                # If self.run_id is None, create a new run; otherwise, use the existing one
                run = self._experiment["mlflow"].start_run(
                    run_id=self.run_id,
                    run_name=self.run_name,
                    nested=self.is_sweep,
                )
                self.run_id = run.info.run_id
                self.mlflow_initialized = (
                    True  # Mark MLflow as initialized successfully
                )

            except Exception as e:
                logger.warning(f"Failed to initialize MLflow: {e}")
                self.mlflow_initialized = False  # Mark MLflow as unavailable

        return self._experiment

    @rank_zero_only
    def log_hyperparams(
        self, params: Dict[str, Union[str, bool, int, float, None]]
    ) -> None:
        """Logs hyperparameter dictionary.

        @type params: Dict[str, Union[str, bool, int, float, None]]
        @param params: Dict of hyperparameters key-value pairs.
        """
        if self.is_tensorboard:
            self.experiment["tensorboard"].add_hparams(
                params,
                {
                    "placeholder_metric": 0
                },  # placeholder metric is needed due to this issue: https://github.com/tensorflow/tensorboard/issues/5476
            )
        if self.is_wandb:
            self.experiment["wandb"].config.update(params)
        if self.is_mlflow:
            self.log_to_mlflow(self.experiment["mlflow"].log_params, params)

    @rank_zero_only
    def log_metric(self, name: str, value: float, step: int) -> None:
        """Logs metric value with name and step.

        @note: step is ommited when logging with wandb to avoid problems
            with inconsistent incrementation.
        @type name: str
        @param name: Metric name
        @type value: float
        @param value: Metric value
        @type step: int
        @param step: Current step
        """
        if self.is_tensorboard:
            self.experiment["tensorboard"].add_scalar(name, value, step)

        if self.is_wandb:
            # let wandb increment step to avoid calls with inconsistent steps
            self.experiment["wandb"].log({name: value})

        if self.is_mlflow:
            self.log_to_mlflow(
                self.experiment["mlflow"].log_metric, name, value, step
            )

    @rank_zero_only
    def log_metrics(self, metrics: Dict[str, float], step: int) -> None:
        """Logs metric dictionary.

        @type metrics: Dict[str, float]
        @param metrics: Dict of metric key-value pairs
        @type step: int
        @param step: Current step
        """
        if self.is_tensorboard:
            for key, value in metrics.items():
                self.experiment["tensorboard"].add_scalar(key, value, step)
        if self.is_wandb:
            self.experiment["wandb"].log(metrics)
        if self.is_mlflow:
            self.log_to_mlflow(
                self.experiment["mlflow"].log_metrics, metrics, step
            )

    @rank_zero_only
    def log_image(self, name: str, img: np.ndarray, step: int) -> None:
        """Logs image with name and step. Note: step is omitted when
        logging with wandb is used to avoid problems with inconsistent
        incrementation.

        @type name: str
        @param name: Caption of the image
        @type img: np.ndarray
        @param img: Image data
        @type step: int
        @param step: Current step
        """
        if self.is_tensorboard:
            self.experiment["tensorboard"].add_image(
                name, img, step, dataformats="HWC"
            )

        if self.is_wandb:
            wandb_image = self.experiment["wandb"].Image(img, caption=name)
            # if step is added here it doesn't work correctly with wandb
            self.experiment["wandb"].log({name: wandb_image})

        if self.is_mlflow:
            # split images into separate directories based on step
            base_path, img_caption = name.rsplit("/", 1)
            img_path = f"{base_path}/{step}/{img_caption}.png"
            self.log_to_mlflow(
                self.experiment["mlflow"].log_image, img, img_path
            )

    @rank_zero_only
    def upload_artifact(
        self,
        path: PathType,
        name: Optional[str] = None,
        typ: str = "artifact",
    ) -> None:
        """Uploads artifact to the logging service.

        @type path: PathType
        @param path: Path to the artifact
        @type name: Optional[str]
        @param name: Name of the artifact, if None then use the name of
            the file
        @type typ: str
        @param typ: Type of the artifact, defaults to "artifact". Only
            used for WandB.
        """
        path = Path(path)
        if self.is_wandb:
            import wandb

            artifact = wandb.Artifact(name=name or path.stem, type=typ)
            artifact.add_file(local_path=str(path))
            artifact.save()

        if self.is_mlflow:
            self.log_to_mlflow(self.upload_artifact_to_mlflow, path, name)

    def upload_artifact_to_mlflow(
        self,
        path: PathType,
        name: Optional[str] = None,
    ) -> None:
        """Uploads artifact specifically to MLflow.

        @type path: PathType
        @param path: Path to the artifact
        @type name: Optional[str]
        @param name: Name of the artifact, if None then use the name of
            the file
        """
        fs = LuxonisFileSystem(
            "mlflow://",
            allow_active_mlflow_run=True,
            allow_local=False,
        )
        fs.put_file(
            local_path=path,
            remote_path=name or path.name,
            mlflow_instance=self.experiment.get("mlflow"),
        )

    @rank_zero_only
    def log_matrix(self, matrix: np.ndarray, name: str, step: int) -> None:
        """Logs matrix to the logging service.

        @type matrix: np.ndarray
        @param matrix: The matrix to log.
        @type name: str
        @param name: The name used to log the matrix.
        @type step: int
        @param step: The current step.
        """
        if self.is_mlflow:
            matrix_data = {
                "flat_array": matrix.flatten().tolist(),
                "shape": matrix.shape,
            }
            self.log_to_mlflow(
                self.experiment["mlflow"].log_dict,
                matrix_data,
                f"{name}.json",
            )

        if self.is_tensorboard:
            matrix_str = np.array2string(matrix, separator=", ")
            self.experiment["tensorboard"].add_text(name, matrix_str, step)

        if self.is_wandb:
            import wandb

            table = wandb.Table(
                columns=["Row Index"]
                + [f"Col {i}" for i in range(matrix.shape[1])]
            )
            for i, row in enumerate(matrix):
                table.add_data(i, *row)
            self.experiment["wandb"].log({f"{name}_table": table}, step=step)

    @rank_zero_only
    def log_images(self, imgs: Dict[str, np.ndarray], step: int) -> None:
        """Logs multiple images.

        @type imgs: Dict[str, np.ndarray]
        @param imgs: Dict of image key-value pairs where key is image
            caption and value is image data
        @type step: int
        @param step: Current step
        """
        for caption, img in imgs.items():
            self.log_image(caption, img, step)

    def _get_next_run_number(self) -> int:
        """Returns number id for next run."""

        log_dirs = [
            path.name
            for path in self.save_directory.iterdir()
            if path.is_dir()
        ]

        nums = [path.split("-")[0] for path in log_dirs]
        nums = [int(num) for num in nums if num.isnumeric()]

        if len(nums) == 0:
            return 0
        return max(nums) + 1

    def close(self) -> None:
        """Finalizes logging and saves unsent logs locally."""
        if self.is_mlflow and any(self.local_logs.values()):
            self.save_logs_locally()

    def _get_run_name(self) -> str:
        """Generates new run name."""
        name_without_number = get_random_name(separator="-", style="lowercase")
        number = self._get_next_run_number()
        return f"{number}-{name_without_number}"

    def _get_latest_run_name(self) -> str:
        """Returns most recently created run name."""
        log_dirs = [
            path.relative_to(self.save_directory).name
            for path in self.save_directory.iterdir()
            if path.is_dir()
        ]
        runs = []
        for ld in log_dirs:
            if ld.split("-")[0].isnumeric():
                runs.append(ld)
        runs.sort(
            key=lambda x: (self.save_directory / x).stat().st_mtime,
            reverse=True,
        )
        return runs[0]



from typing import Dict, Iterable, List, Literal, Optional, Tuple, Union

from typing_extensions import TypedDict

from luxonis_ml.data.utils.constants import LDF_VERSION
from luxonis_ml.utils.pydantic_utils import BaseModelExtraForbid

from .source import LuxonisSource


class Skeletons(TypedDict):
    labels: List[str]
    edges: List[Tuple[int, int]]


class Metadata(BaseModelExtraForbid):
    source: Optional[LuxonisSource]
    ldf_version: str = str(LDF_VERSION)
    classes: Dict[str, Dict[str, int]] = {}
    tasks: Dict[str, List[str]] = {}
    skeletons: Dict[str, Skeletons] = {}
    categorical_encodings: Dict[str, Dict[str, int]] = {}
    metadata_types: Dict[str, Literal["float", "int", "str", "Category"]] = {}
    parent_dataset: Optional[str] = None

    def set_classes(
        self, classes: Union[List[str], Dict[str, int]], task: str
    ) -> None:
        if isinstance(classes, list):
            self.classes[task] = {
                class_name: i
                for i, class_name in enumerate(self._sort_classes(classes))
            }
        else:
            self.classes[task] = classes

    def merge_with(self, other: "Metadata") -> "Metadata":
        """Merge two metadata objects together."""
        if self.ldf_version != other.ldf_version:  # pragma: no cover
            raise ValueError(
                "Cannot merge metadata with different LDF versions"
            )

        merged_classes = {}
        for key in set(self.classes) | set(other.classes):
            if key in self.classes and key in other.classes:
                all_classes = self._sort_classes(
                    set(self.classes[key]) | set(other.classes[key])
                )
                merged_classes[key] = {
                    class_name: i for i, class_name in enumerate(all_classes)
                }
            elif key in self.classes:
                merged_classes[key] = self.classes[key]
            else:
                merged_classes[key] = other.classes[key]

        merged_tasks = {}
        for key in set(self.tasks) | set(other.tasks):
            if key in self.tasks and key in other.tasks:
                merged_tasks[key] = list(
                    set(self.tasks[key] + other.tasks[key])
                )
            elif key in self.tasks:
                merged_tasks[key] = self.tasks[key]
            else:
                merged_tasks[key] = other.tasks[key]

        merged_skeletons = {**self.skeletons, **other.skeletons}

        merged_categorical_encodings = {}
        for key in set(self.categorical_encodings) | set(
            other.categorical_encodings
        ):
            if (
                key in self.categorical_encodings
                and key in other.categorical_encodings
            ):
                merged_inner = self.categorical_encodings[key].copy()
                merged_inner.update(other.categorical_encodings[key])
                merged_categorical_encodings[key] = merged_inner
            elif key in self.categorical_encodings:
                merged_categorical_encodings[key] = self.categorical_encodings[
                    key
                ]
            else:
                merged_categorical_encodings[key] = (
                    other.categorical_encodings[key]
                )

        merged_metadata_types = {**self.metadata_types, **other.metadata_types}
        if self.source is None and other.source is not None:
            merged_source = other.source
        elif self.source is not None and other.source is None:
            merged_source = self.source
        elif self.source is not None and other.source is not None:
            merged_source = self.source.merge_with(other.source)
        else:
            merged_source = None

        return Metadata(
            ldf_version=self.ldf_version,
            source=merged_source,
            classes=merged_classes,
            tasks=merged_tasks,
            skeletons=merged_skeletons,
            categorical_encodings=merged_categorical_encodings,
            metadata_types=merged_metadata_types,  # type: ignore
        )

    def _sort_classes(self, classes: Iterable[str]) -> List[str]:
        return sorted(
            classes,
            key=lambda x: (0, "") if x == "background" else (1, x.lower()),
        )



import random
from typing import Any, Dict, List, Literal, Optional, Tuple, Union

import albumentations as A
import cv2
import numpy as np
from typing_extensions import override

from luxonis_ml.data.augmentations.batch_transform import BatchTransform
from luxonis_ml.data.augmentations.custom import LetterboxResize


class MixUp(BatchTransform):
    def __init__(
        self,
        alpha: Union[float, Tuple[float, float]] = 0.5,
        keep_aspect_ratio: bool = True,
        p: float = 0.5,
    ):
        """MixUp augmentation that merges two images and their
        annotations into one. If images are not of same size then second
        one is first resized to match the first one.

        @type alpha: Union[float, Tuple[float, float]]
        @param alpha: Mixing coefficient, either a single float or a
            tuple representing the range. Defaults to C{0.5}.
        @type keep_aspect_ratio: bool
        @param keep_aspect_ratio: Whether to keep the aspect ratio of
            the second image when resizing. Defaults to C{True}.
        @type p: float, optional
        @param p: Probability of applying the transform. Defaults to
            C{0.5}.
        """
        super().__init__(batch_size=2, p=p)

        self.alpha = (
            alpha if isinstance(alpha, (list, tuple)) else (alpha, alpha)
        )
        if keep_aspect_ratio:
            self.resize_transform = LetterboxResize(1, 1)
        else:
            self.resize_transform = A.Resize(1, 1)

        if not 0 <= self.alpha[0] <= 1 or not 0 <= self.alpha[1] <= 1:
            raise ValueError("Alpha must be in range [0, 1].")

        if self.alpha[0] > self.alpha[1]:
            raise ValueError("Alpha range must be in ascending order.")

    @override
    def apply(
        self,
        image_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        alpha: float,
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of images.

        @type image_batch: List[np.ndarray]
        @param image_batch: Batch of input images to which the
            transformation is applied.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @rtype: List[np.ndarray]
        @return: List of transformed images.
        """

        image1 = image_batch[0]
        image2 = self.resize(image_batch[1], image_shapes, "image", alpha)

        return cv2.addWeighted(image1, alpha, image2, 1 - alpha, 0.0)

    @override
    def apply_to_mask(
        self,
        mask_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        alpha: float,
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of masks.

        Blends masks together. In case of a conflict, the class from the
        mask associated with higher alpha is chosen.

        @type mask_batch: List[np.ndarray]
        @param mask_batch: Batch of input masks to which the
            transformation is applied.
        @type image_shapes: List[Tuple[int, int]]
        @param image_shapes: Shapes of the input images in the batch.
        @type alpha: float
        @param alpha: Mixing coefficient.
        @rtype: List[np.ndarray]
        @return: List of transformed masks.
        """
        mask1, mask2 = mask_batch
        if mask2.size > 0:
            mask2 = self.resize(mask2, image_shapes, "mask")
            if mask2.ndim == 2:
                mask2 = mask2[..., None]

        if mask1.size == 0:
            return mask2
        if mask2.size == 0:
            return mask1

        if alpha >= 0.5:
            mask1[(mask1 == 0) & (mask2 != 0)] = mask2[
                (mask1 == 0) & (mask2 != 0)
            ]
            return mask1

        mask2[(mask2 == 0) & (mask1 != 0)] = mask1[(mask2 == 0) & (mask1 != 0)]
        return mask2

    @override
    def apply_to_instance_mask(
        self,
        mask_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of instance masks.

        @type masks_batch: List[np.ndarray]
        @param masks_batch: Batch of input instance masks to which the
            transformation is applied.
        @rtype: np.ndarray
        @return: Transformed instance masks.
        """
        mask1, mask2 = mask_batch
        if mask2.size > 0:
            mask2 = self.resize(mask2, image_shapes, "mask")
            if mask2.ndim == 2:
                mask2 = mask2[..., None]

        if mask1.size == 0:
            return mask2
        if mask2.size == 0:
            return mask1

        return np.concatenate(mask_batch, axis=-1)

    @override
    def apply_to_bboxes(
        self,
        bboxes_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        rows: int,
        cols: int,
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of bboxes.

        @type bboxes_batch: List[np.ndarray]
        @param bboxes_batch: Batch of input bboxes to which the
            transformation is applied.
        @rtype: np.ndarray
        @return: Transformed bboxes.
        """
        for i in range(len(bboxes_batch)):
            bbox = bboxes_batch[i]
            if bbox.size == 0:  # pragma: no cover
                bboxes_batch[i] = np.zeros((0, 6), dtype=bbox.dtype)

        bboxes_batch[1] = self.resize(
            bboxes_batch[1], image_shapes, "bboxes", rows=rows, cols=cols
        )

        return np.concatenate(bboxes_batch, axis=0)

    @override
    def apply_to_keypoints(
        self,
        keypoints_batch: List[np.ndarray],
        image_shapes: List[Tuple[int, int]],
        **_,
    ) -> np.ndarray:
        """Applies the transformation to a batch of keypoints.

        @type keypoints_batch: List[np.ndarray]
        @param keypoints_batch: Batch of input keypoints to which the
            transformation is applied.
        @rtype: np.ndarray
        @return: Transformed keypoints.
        """
        for i in range(len(keypoints_batch)):
            if keypoints_batch[i].size == 0:  # pragma: no cover
                keypoints_batch[i] = np.zeros(
                    (0, 5), dtype=keypoints_batch[i].dtype
                )

        rows, cols = image_shapes[1]
        keypoints_batch[1] = self.resize(
            keypoints_batch[1],
            image_shapes,
            "keypoints",
            shape=image_shapes[1],
            cols=cols,
            rows=rows,
        )
        return np.concatenate(keypoints_batch, axis=0)

    @override
    def get_params(self) -> Dict[str, Any]:
        """Update parameters.

        @param params: Dictionary containing parameters.
        @type params: Dict[str, Any]
        @return: Dictionary containing updated parameters.
        @rtype: Dict[str, Any]
        """
        alpha = random.uniform(*self.alpha)
        return {"alpha": alpha}

    @override
    def get_params_dependent_on_data(
        self, params: Dict[str, Any], data: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Get parameters dependent on the targets.

        @param params: Dictionary containing parameters.
        @type params: Dict[str, Any]
        @return: Dictionary containing parameters dependent on the
            targets.
        @rtype: Dict[str, Any]
        """
        params = super().get_params_dependent_on_data(params, data)
        image_batch = data["image"]
        return {"image_shapes": [image.shape[:2] for image in image_batch]}

    def resize(
        self,
        data: np.ndarray,
        shapes: List[Tuple[int, int]],
        target_type: Literal["image", "mask", "bboxes", "keypoints"],
        alpha: Optional[float] = None,
        **kwargs,
    ) -> np.ndarray:
        out_height, out_width = shapes[0]
        orig_height, orig_width = shapes[1]
        self.resize_transform.height = out_height
        self.resize_transform.width = out_width
        padding = []
        if isinstance(self.resize_transform, LetterboxResize):
            if alpha is not None:
                self.resize_transform.image_fill_value = (
                    int(255 * alpha),
                    int(255 * alpha),
                    int(255 * alpha),
                )

            padding = LetterboxResize.compute_padding(
                orig_height, orig_width, out_height, out_width
            )

        if target_type == "image":
            return self.resize_transform.apply(data, *padding, **kwargs)
        if target_type == "mask":
            return self.resize_transform.apply_to_mask(
                data, *padding, **kwargs
            )
        if target_type == "bboxes":
            return self.resize_transform.apply_to_bboxes(
                data, *padding, **kwargs
            )
        if target_type == "keypoints":
            return self.resize_transform.apply_to_keypoints(
                data, *padding, **kwargs
            )



from luxonis_ml.guard_extras import guard_missing_extra

with guard_missing_extra("utils"):
    from .config import LuxonisConfig
    from .environ import Environ, environ
    from .filesystem import PUT_FILE_REGISTRY, LuxonisFileSystem
    from .graph import is_acyclic, traverse_graph
    from .logging import deprecated, log_once, setup_logging
    from .pydantic_utils import BaseModelExtraForbid
    from .registry import AutoRegisterMeta, Registry
    from .rich_utils import make_progress_bar


__all__ = [
    "PUT_FILE_REGISTRY",
    "AutoRegisterMeta",
    "BaseModelExtraForbid",
    "Environ",
    "LuxonisConfig",
    "LuxonisFileSystem",
    "Registry",
    "deprecated",
    "environ",
    "is_acyclic",
    "log_once",
    "make_progress_bar",
    "setup_logging",
    "traverse_graph",
]



import subprocess
import sys
import zipfile
from enum import Enum
from importlib.util import find_spec
from pathlib import Path
from typing import (
    Dict,
    Generic,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from loguru import logger

from luxonis_ml.data import DATASETS_REGISTRY, BaseDataset, LuxonisDataset
from luxonis_ml.enums import DatasetType
from luxonis_ml.utils import LuxonisFileSystem, environ
from luxonis_ml.utils.filesystem import _pip_install

from .base_parser import BaseParser
from .classification_directory_parser import ClassificationDirectoryParser
from .coco_parser import COCOParser
from .create_ml_parser import CreateMLParser
from .darknet_parser import DarknetParser
from .native_parser import NativeParser
from .segmentation_mask_directory_parser import SegmentationMaskDirectoryParser
from .solo_parser import SOLOParser
from .tensorflow_csv_parser import TensorflowCSVParser
from .voc_parser import VOCParser
from .yolov4_parser import YoloV4Parser
from .yolov6_parser import YoloV6Parser


class ParserType(Enum):
    DIR = "dir"
    SPLIT = "split"


T = TypeVar("T", str, None)


class LuxonisParser(Generic[T]):
    parsers: Dict[DatasetType, Type[BaseParser]] = {
        DatasetType.COCO: COCOParser,
        DatasetType.VOC: VOCParser,
        DatasetType.DARKNET: DarknetParser,
        DatasetType.YOLOV6: YoloV6Parser,
        DatasetType.YOLOV4: YoloV4Parser,
        DatasetType.CREATEML: CreateMLParser,
        DatasetType.TFCSV: TensorflowCSVParser,
        DatasetType.CLSDIR: ClassificationDirectoryParser,
        DatasetType.SEGMASK: SegmentationMaskDirectoryParser,
        DatasetType.SOLO: SOLOParser,
        DatasetType.NATIVE: NativeParser,
    }

    def __init__(
        self,
        dataset_dir: str,
        *,
        dataset_name: Optional[str] = None,
        save_dir: Optional[Union[Path, str]] = None,
        dataset_plugin: T = None,
        dataset_type: Optional[DatasetType] = None,
        task_name: Optional[Union[str, Dict[str, str]]] = None,
        **kwargs,
    ):
        """High-level abstraction over various parsers.

        Automatically recognizes the dataset format and uses the
        appropriate parser.

        @type dataset_dir: str
        @param dataset_dir: Identifier of the dataset directory.
            Can be one of:
                - Local path to the dataset directory.
                - Remote URL supported by L{LuxonisFileSystem}.
                  - C{gcs://} for Google Cloud Storage
                  - C{s3://} for Amazon S3
                - C{roboflow://} for Roboflow datasets.
                  - Expected format: C{roboflow://workspace/project/version/format}.
            Can be a remote URL supported by L{LuxonisFileSystem}.
        @type dataset_name: Optional[str]
        @param dataset_name: Name of the dataset. If C{None}, the name
            is derived from the name of the dataset directory.
        @type save_dir: Optional[Union[Path, str]]
        @param save_dir: If a remote URL is provided in C{dataset_dir},
            the dataset will be downloaded to this directory. If
            C{None}, the dataset will be downloaded to the current
            working directory.
        @type dataset_plugin: Optional[str]
        @param dataset_plugin: Name of the dataset plugin to use. If
            C{None}, C{LuxonisDataset} is used.
        @type dataset_type: Optional[DatasetType]
        @param dataset_type: If provided, the parser will use this
            dataset type instead of trying to recognize it
            automatically.
        @type task_name: Optional[Union[str, Dict[str, str]]]
        @param task_name: Optional task name(s) for the dataset.
            Can be either a single string, in which case all the records
            added to the dataset will use this value as `task_name`, or
            a dictionary with class names as keys and task names as values.
            In the latter case, the task name for a record with a given
            class name will be taken from the dictionary.
        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} to be passed to the
            constructor of specific L{BaseDataset} implementation.
        """
        save_dir = Path(save_dir) if save_dir else None
        if dataset_dir.startswith("roboflow://"):
            self.dataset_dir, name = self._download_roboflow_dataset(
                dataset_dir, save_dir
            )
        else:
            name = dataset_dir.split("/")[-1]
            local_path = (save_dir or Path.cwd()) / name
            self.dataset_dir = LuxonisFileSystem.download(
                dataset_dir, local_path
            )
        if self.dataset_dir.suffix == ".zip":
            with zipfile.ZipFile(self.dataset_dir, "r") as zip_ref:
                unzip_dir = self.dataset_dir.parent / self.dataset_dir.stem
                logger.info(
                    f"Extracting '{self.dataset_dir.name}' to '{unzip_dir}'"
                )
                zip_ref.extractall(unzip_dir)
                self.dataset_dir = unzip_dir

        if dataset_type:
            self.dataset_type = dataset_type
            self.parser_type = (
                ParserType.DIR
                if Path(self.dataset_dir / "train").exists()
                else ParserType.SPLIT
            )
        else:
            self.dataset_type, self.parser_type = self._recognize_dataset()

        if dataset_plugin is not None:
            self.dataset_constructor = DATASETS_REGISTRY.get(dataset_plugin)
        else:
            self.dataset_constructor = LuxonisDataset

        dataset_name = dataset_name or name.replace(" ", "_").split(".")[0]

        self.dataset = self.dataset_constructor(
            dataset_name=dataset_name,  # type: ignore
            **kwargs,
        )
        self.parser = self.parsers[self.dataset_type](
            self.dataset, self.dataset_type, task_name
        )

    @overload
    def parse(self: "LuxonisParser[str]", **kwargs) -> BaseDataset: ...

    @overload
    def parse(self: "LuxonisParser[None]", **kwargs) -> LuxonisDataset: ...

    def parse(self, **kwargs) -> BaseDataset:
        """Parses the dataset and returns it in LuxonisDataset format.

        If the dataset already exists, parsing will be skipped and the
        existing dataset will be returned instead.

        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} for specific parser
            implementation.
        @rtype: LuxonisDataset
        @return: Parsed dataset in L{LuxonisDataset} format.
        """
        if self.parser_type == ParserType.DIR:
            dataset = self._parse_dir(**kwargs)
        else:
            dataset = self._parse_split(**kwargs)

        logger.info("Dataset parsed successfully.")
        return dataset

    def _recognize_dataset(self) -> Tuple[DatasetType, ParserType]:
        """Recognizes the dataset format and parser type.

        @rtype: Tuple[DatasetType, ParserType]
        @return: Tuple of dataset type and parser type.
        """
        for typ, parser in self.parsers.items():
            if parser.validate(self.dataset_dir):
                logger.info(f"Recognized dataset format as <{typ.name}>")
                return typ, ParserType.DIR

        for typ, parser in self.parsers.items():
            if parser.validate_split(self.dataset_dir):
                logger.info(
                    f"Recognized dataset format as a split of <{typ.name}>"
                )
                return typ, ParserType.SPLIT

        raise ValueError(
            f"Dataset {self.dataset_dir} is not in expected format for any of the parsers."
        )

    def _parse_dir(self, **kwargs) -> BaseDataset:
        """Parses all present data in LuxonisDataset format.

        Check under each parser for the expected directory structure.

        @type kwargs: Dict[str, Any]
        @param kwargs: Additional C{kwargs} for specific parser
            function.
        @rtype: LuxonisDataset
        @return: C{LDF} with all the images and annotations parsed.
        """

        return self.parser.parse_dir(self.dataset_dir, **kwargs)

    def _parse_split(
        self,
        split: Optional[str] = None,
        random_split: bool = True,
        split_ratios: Optional[Dict[str, float]] = None,
        **kwargs,
    ) -> BaseDataset:
        """Parses data from a subdirectory representing a single split.

        Should be used if adding/changing only specific split. Check
        under each parser for expected directory structure.

        @type split: Optional[Literal["train", "val", "test"]]
        @param split: As what split the data will be added to LDF. If
            set, C{split_ratios} and C{random_split} are ignored.
        @type random_split: bool
        @param random_split: If random splits should be made. If
            C{True}, C{split_ratios} are used.
        @type split_ratios: Optional[Dict[str, float]]
        @param split_ratios: Ratios for random splits. Only used if
            C{random_split} is C{True}. Defaults to C{{"train": 0.8,
            "val": 0.1, "test": 0.1}}.
        @type kwargs: Dict[str, Any]
        @param kwargs: Additional kwargs for specific parser
            implementation.
        @rtype: LuxonisDataset
        @return: C{LDF} with all the images and annotations parsed.
        """
        parsed_kwargs = self.parser.validate_split(self.dataset_dir)
        if parsed_kwargs is None:
            raise ValueError(
                f"Dataset {self.dataset_dir} is not in the expected format for {self.dataset_type} parser."
            )
        return self.parser.parse_split(
            split, random_split, split_ratios, **parsed_kwargs, **kwargs
        )

    def _download_roboflow_dataset(
        self, dataset_dir: str, local_path: Optional[Path]
    ) -> Tuple[Path, str]:
        if find_spec("roboflow") is None:  # pragma: no cover
            _pip_install("roboflow", "roboflow~=1.1.0")
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "uninstall",
                    "-y",
                    "opencv-python",
                    "opencv-python-headless",
                ],
                check=False,
            )
            subprocess.run(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "opencv-python~=4.10.0",
                ],
                check=False,
            )

        from roboflow import Roboflow

        if environ.ROBOFLOW_API_KEY is None:
            raise RuntimeError(
                "ROBOFLOW_API_KEY environment variable is not set. "
                "Please set it to your Roboflow API key."
            )

        rf = Roboflow(api_key=environ.ROBOFLOW_API_KEY)
        parts = dataset_dir.split("roboflow://")[1].split("/")
        if len(parts) != 4:
            raise ValueError(
                f"Incorrect Roboflow dataset URL: `{dataset_dir}`. "
                "Expected format: `roboflow://workspace/project/version/format`."
            )
        workspace, project, version, format = dataset_dir.split("roboflow://")[
            1
        ].split("/")
        try:
            version = int(version)
        except ValueError as e:
            raise ValueError(
                f"Roboflow version must be an integer, got `{version}`."
            ) from e

        local_path = local_path or Path.cwd() / f"{project}_{format}"
        dataset = (
            rf.workspace(workspace)
            .project(project)
            .version(int(version))
            .download(format, str(local_path / project))
        )
        return Path(dataset.location), project
